#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from builtins import super
import logging
from holado_core.common.tools.tools import Tools
from holado.common.context.session_context import SessionContext
import threading
from holado_test.test_config import TestConfig
import os
import time

logger = logging.getLogger(__name__)


class BehaveSessionContext(SessionContext):
    
    def __init__(self, name="Session"):
        super().__init__(name)
        
        self.__multitask_step_lock = threading.RLock()

    @classmethod
    def has_instance_scenario_context(cls):
        return hasattr(SessionContext.instance(), "has_scenario_context") and SessionContext.instance().has_scenario_context()

    def before_all(self, behave_context):
        log_prefix = f"[Before session] "
        
        with self._multitask_lock:
            self.behave_manager.set_main_context(behave_context)
            self.report_manager.before_all()
        
        # Set variable with session context instance
        self.variable_manager.register_variable("SESSION_CONTEXT", self)
        
        logger.info(f"{log_prefix}Doing all post processes of a previous session if needed...")
        if TestConfig.do_previous_post_processes:
            self.do_all_persisted_post_processes()
        
        # Abort execution if SUT is not operational
        # Note: it must be done after post processes, since a post process can restore SUT
        self.manage_execution_abort()
        
        
    def after_all(self):
        log_prefix = f"[After session] "
            
        # Post processes
        logger.info(f"{log_prefix}Post processing...")
        self.do_post_processes()
        
        with self._multitask_lock:
            self.report_manager.after_all()
            self.behave_manager.clear()

    def before_feature(self, feature):
        from holado_system.system.global_system import GlobalSystem
        from holado_helper.debug.memory.memory_profiler import MemoryProfiler
        from holado_test.common.context.feature_context import FeatureContext
        
        if self.behave_manager.is_execution_aborted():
            return
        
        log_prefix = f"[Before feature '{feature.name}'] "
        
        with self._multitask_lock:
            # Logs
            logger.info("="*150)
            logger.info(f"Feature [{feature.name}]")
            
            logger.info(f"{log_prefix}Begin")
            if self.has_feature_context(is_reference=True):
                from holado_core.common.exceptions.technical_exception import TechnicalException
                raise TechnicalException(f"{log_prefix}A feature context is already defined")
            
            GlobalSystem.log_resource_usage(prefix=log_prefix, level=logging.INFO, logger_=logger)
            if TestConfig.profile_memory_in_features and MemoryProfiler.is_tracker_available():
                MemoryProfiler.create_or_reset_tracker_of_objects_summary_changes("features summary")
                # MemoryProfiler.create_or_reset_tracker_of_objects_changes("features objects")
                
            # Feature context
            feature_context = FeatureContext(feature)
            self.__set_feature_context(feature_context)
            
            # Set variable with feature context instance
            self.variable_manager.register_variable("FEATURE_CONTEXT", feature_context)
            
            # Report
            try:
                self.report_manager.before_feature(feature_context, feature)
            except:
                logger.exception(f"{log_prefix}Error while updating report before feature")
            
            logger.info(f"{log_prefix}End")
        
    def after_feature(self, feature):
        from holado_system.system.global_system import GlobalSystem
        from holado_helper.debug.memory.memory_profiler import MemoryProfiler
        
        if self.behave_manager.is_execution_aborted():
            return
        
        log_prefix = f"[After feature '{feature.name}'] "
        
        with self._multitask_lock:
            logger.info(f"{log_prefix}Begin")
            if not self.has_feature_context(is_reference=True):
                from holado_core.common.exceptions.technical_exception import TechnicalException
                raise TechnicalException(f"{log_prefix}No feature context is defined")
            
            if TestConfig.profile_memory_in_features and MemoryProfiler.is_tracker_available():
                MemoryProfiler.log_tracker_diff(name="features summary", prefix=log_prefix, level=logging.INFO, logger_=logger)  # @UndefinedVariable
                # MemoryProfiler.log_tracker_diff(name="features objects", prefix="[After feature] ", level=logging.INFO, logger_=logger)  # @UndefinedVariable
            GlobalSystem.log_resource_usage(prefix=log_prefix, level=logging.INFO, logger_=logger)
            
            # End feature context
            self.get_feature_context().end()
            
            # Report
            try:
                self.report_manager.after_feature(feature)
            except:
                logger.exception(f"{log_prefix}Error while updating report after feature")
            
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"{log_prefix}Deleting feature context")
            self.__delete_feature_context()
            
            logger.info(f"{log_prefix}End")
            logger.info(f"Finished feature [{feature.name}]")
        
    def has_feature_context(self, is_reference=None, do_log=False):
        return self.multitask_manager.has_feature_context(is_reference=is_reference, do_log=do_log)
    
    def get_feature_context(self):
        return self.multitask_manager.get_feature_context()
    
    def __set_feature_context(self, feature_context):
        return self.multitask_manager.set_feature_context(feature_context)
    
    def __delete_feature_context(self):
        return self.multitask_manager.delete_feature_context()
    
    def before_scenario(self, scenario):
        from holado_system.system.global_system import GlobalSystem
        from holado_helper.debug.memory.memory_profiler import MemoryProfiler
        from holado_test.common.context.scenario_context import ScenarioContext
        
        if self.behave_manager.is_execution_aborted():
            return
        
        log_prefix = f"[Before scenario '{scenario.name}'] "
        
        with self._multitask_lock:
            logger.info("-"*150)
            logger.info(f"Scenario [{scenario.name}]")
            
            logger.info(f"{log_prefix}Begin")
            if not self.has_feature_context(is_reference=True):
                from holado_core.common.exceptions.technical_exception import TechnicalException
                raise TechnicalException(f"{log_prefix}No feature context is defined")
            
            # Create and initialize ScenarioContext
            scenario_context = ScenarioContext(scenario)
            self.get_feature_context().add_scenario(scenario_context)
            
            # Report: create scenario report
            try:
                self.report_manager.before_scenario(scenario_context, scenario)
            except:
                logger.exception(f"{log_prefix}Error while updating report before scenario")
            
            # Set variable with scenario context instance
            # Note: must be after scenario report creation
            self.get_scenario_context().get_variable_manager().register_variable("SCENARIO_CONTEXT", self.get_scenario_context())
            
            # Behave context
            try:
                self.behave_manager.before_scenario()
            except:
                logger.exception(f"{log_prefix}Error while updating behave context before scenario")
            
            GlobalSystem.log_resource_usage(log_prefix, level=logging.INFO, logger_=logger)
            if TestConfig.profile_memory_in_scenarios and MemoryProfiler.is_tracker_available():
                MemoryProfiler.create_or_reset_tracker_of_objects_summary_changes("scenarios summary")
                # MemoryProfiler.create_or_reset_tracker_of_objects_changes("scenarios objects")
            
            if TestConfig.do_previous_post_processes:
                logger.info(f"{log_prefix}Doing previous scenario post processes if needed...")
                self.get_scenario_context().do_persisted_post_processes()
            
            logger.info(f"{log_prefix}End")
            logger.info(f"Start scenario [{scenario.name}]")
        
    def get_SUT_status(self):
        """ Return if SUT is operational and error message
        Override it in a subclass to activate this feature.
        """
        return True, None

    def manage_execution_abort(self, scenario=None):
        """ Manage cases execution must be aborted for outside reasons, after execution of post processes.
        It uses the return of get_SUT_status to abort if SUT is not operational.
        """
        if TestConfig.manage_execution_abort:
            is_sut_ok, sut_error = self.get_SUT_status()
            if not is_sut_ok:
                if scenario:
                    scenario.sut_failed = True
                    scenario.sut_error = sut_error
                self.abort_execution(reason=f"SUT is not operational: {sut_error}")

    def manage_execution_abort_on_failure(self, scenario):
        """ Abort execution on scenario failure, before post processes.
        Note: post processes won't be executed.
        Usage: Set environment variable HOLADO_TEST_ABORT_EXECUTION_ON_FAILURE=True to activate this feature.
               It is usefull for debug of SUT and/or testing solution, when leaving whole as is on failure.
        """
        if TestConfig.abort_execution_on_failure and scenario.status.has_failed():
            self.abort_execution(reason=f"Abort after scenario failure")

    def abort_execution(self, reason):
        """ Abort execution.
        """
        self.behave_manager.abort_execution(reason=reason)
        
        # Logs
        logger.print(f"ABORT EXECUTION: {reason}")
        print(f"ABORT EXECUTION: {reason}")
        
        # Create abort report
        self.report_manager.report_abort(reason)
    
    def after_scenario(self, scenario):
        from holado_core.common.exceptions.technical_exception import TechnicalException
        from holado_system.system.global_system import GlobalSystem
        from holado_helper.debug.memory.memory_profiler import MemoryProfiler

        # Abort on failure if configured to
        self.manage_execution_abort_on_failure(scenario)

        # Don't process after scenario if execution is aborted
        if self.behave_manager.is_execution_aborted():
            return

        log_prefix = f"[After scenario '{scenario.name}'] "

        with self._multitask_lock:
            # End scenario
            if not self.has_feature_context(is_reference=True):
                raise TechnicalException(f"{log_prefix}No feature context is defined")
            if not self.has_scenario_context(is_reference=True):
                raise TechnicalException(f"{log_prefix}No scenario context is defined")
            self.get_scenario_context().end()

            # Process actions at scenario end
            try:
                self.after_scenario_end(scenario)
            except:
                logger.exception(f"{log_prefix}Error while processing actions after scenario end")

            # Process after scenario
            logger.info(f"{log_prefix}Begin")
            # logger.printf"++++++++++++ scenario: {Tools.represent_object(scenario)}")

            logger.info(f"{log_prefix}Resource usage:")
            if TestConfig.profile_memory_in_scenarios and MemoryProfiler.is_tracker_available():
                MemoryProfiler.log_tracker_diff(name="scenarios summary", prefix=log_prefix, level=logging.INFO, logger_=logger)  # @UndefinedVariable
                # MemoryProfiler.log_tracker_diff(name="scenarios objects", prefix="[After scenario] ", level=logging.INFO, logger_=logger)  # @UndefinedVariable
            GlobalSystem.log_resource_usage(prefix=log_prefix, level=logging.INFO, logger_=logger)

            # Post processes
            logger.info(f"{log_prefix}Post processing...")
            self.get_scenario_context().scope_manager.reset_scope_level()
            self.get_scenario_context().do_post_processes()

            # Abort execution if SUT has fall not operational during scenario
            # Note: it must be done after post processes (a post process can restore SUT) and before reports generation (to manage status 'Failed in SUT' in reports)
            self.manage_execution_abort(scenario)

            # Report
            logger.info(f"{log_prefix}Generating reports...")
            try:
                self.report_manager.after_scenario(scenario)
            except:
                logger.exception(f"{log_prefix}Error while updating report after scenario")

            # Delete scenario context
            logger.info(f"{log_prefix}Deleting scenario context...")
            # if Tools.do_log(logger, logging.DEBUG):
            #     logger.debug("Deleting context of scenario [{}]".format(scenario.name))
            self.__delete_scenario_context()

            # Remove all threads
            logger.info(f"{log_prefix}Interrupting and unregistering all threads...")
            # if Tools.do_log(logger, logging.DEBUG):
            #     logger.debug("Interrupting and unregistering all threads for scenario [{}]".format(scenario.name))
            #TODO: For case of multiple scenarios launched in parallel, interrupt only threads related to current scenario (scenario launched by this thread)
            self.threads_manager.interrupt_all_threads(scope="Scenario")
            self.threads_manager.unregister_all_threads(scope="Scenario", keep_alive=False)

            logger.info(f"{log_prefix}End")
            logger.info(f"Finished scenario [{scenario.name}]")
    
    def after_scenario_end(self, scenario):
        from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
        from holado_test.common.context.context_tools import ContextTools
        from holado_test.behave.scenario.behave_scenario_tools import BehaveScenarioTools
        
        # Log error on failing scenario
        status_info = BehaveScenarioTools.get_current_scenario_status_info(scenario)
        has_failed = status_info.validation_status != "Passed"
        if has_failed:
            msg_list = []
            category_str = f" => {status_info.validation_category}" if status_info.validation_category else ""
            msg_list.append(f"Scenario {status_info.validation_status}{category_str}: {ContextTools.format_context_period(status_info.scenario_context)} {BehaveScenarioTools.format_scenario_short_description(scenario)} - {BehaveStepTools.format_step_short_description(status_info.step_failed, status_info.step_failed_nb, step_context=status_info.step_context, dt_ref=status_info.scenario_context.start_datetime, has_failed=has_failed)}")
            step_error_message = BehaveStepTools.get_step_error_message(status_info.step_failed)
            if step_error_message:
                msg_list.append(step_error_message)
            msg = "\n".join(msg_list)
            logger.error(msg)

            wait_s = os.getenv("HOLADO_TEST_WAIT_AFTER_FAILED_SCENARIO_SECONDS", None)
            if wait_s is not None:
                logger.warning(f"Waiting {wait_s} seconds on failed scenario")
                time.sleep(int(wait_s))
                logger.warning(f"Finished to wait {wait_s} seconds on failed scenario")

    def has_scenario_context(self, is_reference=None):
        return self.has_feature_context(is_reference=is_reference) and self.get_feature_context().has_scenario
    
    def get_scenario_context(self):
        return self.get_feature_context().current_scenario
    
    def __delete_scenario_context(self):
        self.get_scenario_context().delete_object()
        
    def before_step(self, step):
        log_prefix = f"[Before step '{step}'] "
        
        with self.__multitask_step_lock:
            from holado_test.behave.context.behave_step_context import BehaveStepContext
            from holado_core.common.exceptions.technical_exception import TechnicalException
            
            if not self.has_feature_context(is_reference=None):
                # Look again but do logs before raising exception
                self.has_feature_context(is_reference=None, do_log=True)
                raise TechnicalException(f"{log_prefix}No feature context is defined (step: {step})")
            if not self.has_scenario_context(is_reference=None):
                raise TechnicalException(f"{log_prefix}No scenario context is defined (step: {step})")
            scenario_context = self.get_scenario_context()
            
            # Manage step context
            step_context = BehaveStepContext(step)
            scenario_context.scope_manager.set_step_context(step_context)
            
            # Update scenario context
            step_level = scenario_context.scope_manager.scope_level("steps")
            if step_level == 0:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"{log_prefix}Add step {step_context} in scenario {scenario_context}")
                scenario_context.add_step(step_context)
            
            # Manage scope in define
            if scenario_context.block_manager.is_in_define \
                    and (scenario_context.block_manager.is_in_sub_define         # needed to add in define the "end define" steps of sub-define \
                         or step.name not in ["end for", "end while", "end function"]):     # needed to not add in define the "end define" steps of current define
                from holado_test.behave.behave import format_step
                step_str = format_step(step)
                scenario_context.block_manager.scope_in_define.add_steps(step_str)
                
                # Set step status
                step_context.status = "defined"
            
            # Report
            try:
                self.report_manager.before_step(step_context, step, step_level)
            except:
                logger.exception(f"{log_prefix}Error while updating report before step")
    
    def after_step(self, step, has_started=True):
        """Process after step
        @param step: step instance
        @param has_started: if False, the step is added but without execution. 
            It is usually True and before_step was called before, except for undefined and skipped steps.
        """
        log_prefix = f"[After step '{step}'] "
        
        with self.__multitask_step_lock:
            from holado_test.common.context.step_context import StepContext
            from holado_core.common.exceptions.technical_exception import TechnicalException
            
            if not self.has_feature_context(is_reference=None):
                raise TechnicalException(f"{log_prefix}No feature context is defined (step: {step})")
            if not self.has_scenario_context(is_reference=None):
                raise TechnicalException(f"{log_prefix}No scenario context is defined (step: {step})")
            scenario_context = self.get_scenario_context()
            
            # Manage step context
            if has_started:
                step_context = scenario_context.scope_manager.get_step_context()
                step_context.end()
                if step_context.status is None:
                    step_context.status = step.status.name
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"{log_prefix}Ended step {step_context} in scenario {scenario_context}")
            else:
                # Manage step context
                step_context = StepContext(step, do_start=False)
                scenario_context.scope_manager.set_step_context(step_context)
                
                # Update scenario context
                step_level = scenario_context.scope_manager.scope_level("steps")
                if step_level == 0:
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug(f"{log_prefix}Add step {step_context} in scenario {scenario_context}")
                    scenario_context.add_step(step_context)
                
                # Process before step in reports
                try:
                    self.report_manager.before_step(step_context, step, step_level)
                except:
                    logger.exception(f"{log_prefix}Error while updating report before step for unstarted step")
            
            # Report
            try:
                step_level = scenario_context.scope_manager.scope_level("steps")
                self.report_manager.after_step(step_context, step, step_level)
            except:
                logger.exception(f"{log_prefix}Error while updating report after step")
    
    def has_step_context(self):
        return self.has_feature_context() and self.get_feature_context().has_scenario and self.get_scenario_context().has_step()
    
    def get_step_context(self):
        return self.get_scenario_context().get_current_step()
        







#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.tools.string_tools import StrTools

logger = logging.getLogger(__name__)


class Binary():
    """
    Tools to manipulate binary data.
    """
    
    @classmethod
    def convert_bin_str_to_hex_str(cls, bin_str, right_padding=False, nbbits_by_block=8):
        """
        Convert binary string to hexadecimal string (ex: 'A2FF').
        
        If binary string has a length that is not a multiple of nbbits_by_block, a padding with zeros is made at left or right, depending on parameter right_padding.
        """
        # Manage padding for bits out of blocks of nbbits_by_block bits
        nb_bits_out_of_block = len(bin_str) % nbbits_by_block
        if nb_bits_out_of_block > 0:
            padding_bits = "0" * (nbbits_by_block - nb_bits_out_of_block)
            if right_padding:
                bin_str += padding_bits
            else:
                bin_str = padding_bits + bin_str
        if len(bin_str) % nbbits_by_block != 0:
            raise TechnicalException("Failed to pad binary series")
        
        # Manage 4-bits blocks
        res_list = []
        for i in range(len(bin_str) // 4):
            res_list.append("{:X}".format(int(bin_str[i*4:i*4+4],2)))
        return "".join(res_list)
    
    @classmethod
    def convert_hex_str_to_bin_str(cls, hex_str, bit_length=None, right_padded=False):
        """
        Convert hexadecimal string to binary string (ex: '10001101').
        """
        res = "".join(["{:04b}".format(int(hc, 16)) for hc in hex_str])
        
        if bit_length is not None:
            if bit_length > len(hex_str) * 4:
                raise FunctionalException(f"Data bit length ({bit_length}) is greater than data size ({len(hex_str) * 4})")
            if right_padded:
                res = res[:bit_length]
            else:
                res = res[-bit_length:]
        
        return res
        
    @classmethod
    def pad_bin_str_with_itself(cls, bin_str, bit_length, right_padding=False):
        if len(bin_str) == 0:
            raise FunctionalException(f"Binary string is empty, it is not possible to pad it with itself")
        if bit_length < len(bin_str):
            raise FunctionalException(f"Binary string has a length ({len(bin_str)}) greater than expected length ({bit_length}) after padding")
        
        res = bin_str
        while len(res) < bit_length:
            missing_len = bit_length - len(res)
            if len(bin_str) <= missing_len:
                pad_str = bin_str
            else:
                pad_str = bin_str[:missing_len]
            if right_padding:
                res += pad_str
            else:
                res = pad_str + res
        
        return res

    @classmethod
    def pad_data(cls, data_str, data_length, padding_char='0', right_padding=False):
        if data_length < len(data_str):
            raise FunctionalException(f"Data has a length ({len(data_str)}) greater than expected length ({data_length}) after padding")
        
        pad_str = padding_char * (data_length - len(data_str))
        if right_padding:
            return data_str + pad_str
        else:
            return pad_str + data_str

    @classmethod
    def data_length_after_padding(cls, data_length, multiple_of_length):
        if data_length == 0:
            return 0
        else:
            return ((data_length - 1) // multiple_of_length + 1) * multiple_of_length

    @classmethod
    def get_data_bit_length(cls, data):
        """Return number of bits of data after conversion in binary format.
        Note: for string data type, data is supposed to be in hexadecimal format.
        """
        if isinstance(data, int):
            return len(f"{data:b}")
        elif isinstance(data, bytes):
            return len(data) * 8
        elif isinstance(data, str):
            if StrTools.is_hex(data):
                return len(data) * 4
            else:
                raise TechnicalException(f"String data must be in hexadecimal format: [{data}]")
        else:
            raise TechnicalException(f"Unmanaged data of type {type(data)}")






# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


from holado_test.scenario.step_tools import StepTools
from holado.common.context.session_context import SessionContext
from holado_test.behave.behave import *  # @UnusedWildImport
import logging
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_value.common.tables.converters.value_table_converter import ValueTableConverter
import email.message
from holado_value.common.tables.value_table_manager import ValueTableManager
from holado_core.common.exceptions.technical_exception import TechnicalException

logger = logging.getLogger(__name__)



def __get_scenario_context():
    return SessionContext.instance().get_scenario_context()

def __get_variable_manager():
    return __get_scenario_context().get_variable_manager()


@Step(r"(?P<var_name>{Variable}) = new email message")
def step_impl(context, var_name):
    """ Create an email message with standard library email
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    table = BehaveStepTools.get_step_table(context)
    params = ValueTableConverter.convert_name_value_table_2_dict(table)
    
    policy = params.pop('policy', None)
    res = email.message.EmailMessage(policy=policy)
    
    for key, value in params.items():
        res[key] = value
    
    __get_variable_manager().register_variable(var_name, res)

@Step(r"(?P<var_name>{Variable}) = convert email (?P<mail>{Variable}) to name/value table")
def step_impl(context, var_name, mail):
    """ Convert email (header content and body) to a name/value table
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    mail = StepTools.evaluate_variable_value(mail)
    
    if mail.is_multipart():
        raise TechnicalException(f"Unable to convert multipart mail to name/value table")
    
    mail_dict = {item[0]:str(item[1]) for item in mail.items()}
    mail_dict['Body'] = mail.get_body().get_content()
    res = ValueTableManager.convert_dict_2_name_value_table(mail_dict)
    
    __get_variable_manager().register_variable(var_name, res)







#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import copy
from holado_core.common.inspectors.tools.inspect_context import InspectContext
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_ui.ui.handlers.ui_context import UIContext


class UIInspectContext(InspectContext):
    """ UI inspect context
    """
    
    __instance_default = None
    
    def __init__(self, find_type=None, inspect_context=None, ui_context=None):
        super().__init__(find_type, inspect_context)
        self.__ui_context = ui_context
    
    @property
    def ui_context(self) -> UIContext:
        """
        @return UI context
        """
        return self.__ui_context
    
    @ui_context.setter
    def ui_context(self, ui_context:UIContext):
        self.__ui_context = ui_context
    
    def with_ui_context(self, ui_context):
        """
        Note: if wanted UI context is the same, self instance is returned, else a new one is returned
        @param ui_context UI context
        @return Same context but with given UI context
        """
        if ui_context == self.ui_context:
            return self
        else:
            res = copy.deepcopy(self)
            res.ui_context = ui_context
            return res

    def without_ui_context(self):
        """
        @return Same context but without UI context
        """
        return self.with_ui_context(self, None)
    
    def with_previous_ui_context(self):
        """
        @return Same context but with previous UI context
        """
        if self.ui_context is not None:
            return self.with_ui_context(self.__ui_context.get_previous_context())
        else:
            raise TechnicalException("UI context is not defined")
    
    @staticmethod
    def default():
        """
        @return Default UI inspect context
        """
        if InspectContext.__instance_default is None:
            InspectContext.__instance_default = InspectContext()
        return InspectContext.__instance_default
    
    
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import abc
import time
from holado_multitask.multithreading.functionthreaded import FunctionThreaded
from holado_core.common.tools.tools import Tools
from holado_core.common.exceptions.timeout_exception import TimeoutException
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.handlers.exceptions.redo_exceptions import RedoIgnoredException,\
    RedoStopRetryException
from holado.holado_config import Config
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


class BaseRedo(object):
    """ Base class managing redo of a process until timeout.
    """
    __metaclass__ = abc.ABCMeta

    """
    Default timeout.
    """
    DEFAULT_TIMEOUT_S = 30
    """
    Default timeout for each process.
    """
    DEFAULT_PROCESS_TIMEOUT_S = None
    """
    Default polling delay between each process.
    """
    DEFAULT_POLLING_INTERVAL_S = 0.1
    """
    Default number of allowed retries.
    """
    DEFAULT_NB_RETRIES = 0
    """
    Number of successive failures that must appear before throwing an error.
    """
    DEFAULT_NB_SUCCESSIVE_FAILURES_BEFORE_ERROR = 1
    """
    Default waiting interval after a failure and before a new retry.
    """
    DEFAULT_WAITING_INTERVAL_AFTER_FAILURE_S = 0.1


    def __init__(self, name):
        """
        @param name Name describing redo
        """
        self.__name = name
        
        self.__timeout = BaseRedo.DEFAULT_TIMEOUT_S
        self.__process_timeout = BaseRedo.DEFAULT_PROCESS_TIMEOUT_S
        self.__accepted_time = None
        self.__interval = BaseRedo.DEFAULT_POLLING_INTERVAL_S
        self.__interval_after_failure = BaseRedo.DEFAULT_WAITING_INTERVAL_AFTER_FAILURE_S
        self.__nb_retries = BaseRedo.DEFAULT_NB_RETRIES
        self.__nb_successive_failures_before_error = BaseRedo.DEFAULT_NB_SUCCESSIVE_FAILURES_BEFORE_ERROR
        
        self.__ignored_exceptions = [RedoIgnoredException]
        self.__stop_retry_exceptions = [RedoStopRetryException]
        self.__do_process_in_thread = True
        self.__do_raise_on_stop_retry = True
        self.__func_to_call_on_timeout = None       # Specify function to call on timeout
        self.__do_raise_on_timeout = True           # Specify if a TimeoutException must be raised on timeout
        self.__do_raise_last_exception = True       # Specify if last exception must be raised after timeout, whereas it is configured to not raise a timeout exception
        
        self.__run_counter = 0
        self.__interrupt_counter = 0
        self.__failure_counter = 0
        
        self.__current_process_result = None
        self.__current_process_exception = None
        self.__previous_process_exception_stack_trace = None
        self.__previous_process_exception_counter = 0
        
        self.__start_time = None
        self.__end_time = None
    
    @property
    def name(self):
        """
        @return redo name
        """
        return self.__name
    
    @property
    def process_timeout(self):
        """
        @return The process timeout in seconds
        """
        if self.__process_timeout is not None:
            return self.__process_timeout
        else:
            return self.timeout
    
    @property
    def timeout(self):
        """
        @return The timeout in seconds
        """
        return self.__timeout
    
    @property
    def accepted_time(self):
        """
        @return The accepted spent time
        """
        return self.__accepted_time
    
    @property
    def run_counter(self):
        return self.__run_counter
    
    @property
    def start_time(self):
        return self.__start_time
    
    @property
    def end_time(self):
        return self.__end_time
    
    @property
    def spent_time(self):
        if self.start_time is None:
            return None
        elif self.__end_time is None:
            return Tools.timer_s() - self.start_time
        else:
            return self.__end_time - self.start_time
    
    def _process(self):
        """
        @return Result of a run
        """
        raise NotImplementedError

    def _process_interrupt(self, thread):
        """
        Method launched after a run interruption
        Note: for process implementation that launch uninterruptible code, self method can be used to stop the thread by calling thread.stop()
        @param thread Thread running the process method
        """
        # Nothing by default
        pass
    
    def _is_redo_needed(self, result):
        """
        @param result Current run result.
        @return True if a redo is needed according given result.  
        """
        raise NotImplementedError

    def execute(self):
        """
        @return Result of execution
        """
        if self.__do_process_in_thread:
            return self.__execute_with_thread()
        else:
            return self.__execute_without_thread()

    def __execute_without_thread(self):
        """
        @return Result of execution without executing process in a thread
        """
        result = None
        last_exception = None
        stop_retry = False
        self.__start_time = Tools.timer_s()
        
        try:
            while not stop_retry:
                try:
                    self.__run_counter += 1
                    if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
                        logger.trace(f"[{self.name} - {self.__run_counter}] Processing run {self.__run_counter}")
                    
                    self.__execute_process()
                    
                    if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
                        logger.trace(f"[{self.name} - {self.__run_counter}] Post processing")
                    do_return, result = self.__post_process(interrupted=False)
                    if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
                        logger.trace(f"[{self.name} - {self.__run_counter}] do return: {do_return} ; result: [{result}] (type: {Typing.get_object_class_fullname(result)})")
                except Exception as exc:
                    has_failure, stop_retry = self.__manage_process_exception(exc, last_exception)
                    last_exception = exc
                else:
                    has_failure = False
                    self.__failure_counter = 0
                    if do_return: # Expected result has appeared
                        # Verify result doesn't appeared too late compared to accepted time
                        self.__raise_accepted_time_if_needed()
                        # Return obtained result
                        return result
                    
                if not stop_retry:
                    stop_retry = self.__manage_process_timeout(last_exception)
                    if stop_retry:
                        # Timeout is reached, but it is configured to not raise an exception ; return last polled result
                        return result
                    self.__wait_after_process(has_failure)
        finally:
            self.__end_time = Tools.timer_s()
            if self.run_counter > 1 or self.__interrupt_counter > 0:
                logger.info(f"redo [{self.name}] has made {self.run_counter} runs and {self.__interrupt_counter} interrupts (process timeout: {self.process_timeout} s)")
        
        return result

    def __execute_with_thread(self):
        """
        @return Result of execution with process execution in a thread
        """
        result = None
        last_exception = None
        stop_retry = False
        self.__start_time = Tools.timer_s()
        
        try:
            while not stop_retry:
                # Create new thread to process
                thread = self.__get_execute_thread()
                
                try:
                    self.__run_counter += 1
                    if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
                        logger.trace(f"[{self.name} - {self.__run_counter}] Processing run {self.__run_counter}")
                    
                    thread.start()
                    self.__wait_process(thread)
                    interrupted = self.__interrupt_thread_if_still_alive(thread)
                    
                    if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
                        logger.trace(f"[{self.name} - {self.__run_counter}] Post processing (interrupted: {interrupted})")
                    do_return, result = self.__post_process(interrupted)
                    if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
                        logger.trace(f"[{self.name} - {self.__run_counter}] do return: {do_return} ; result: [{result}] (type: {Typing.get_object_class_fullname(result)})")
                except Exception as exc:
                    has_failure, stop_retry = self.__manage_process_exception(exc, last_exception)
                    last_exception = exc
                else:
                    has_failure = False
                    self.__failure_counter = 0
                    if do_return: # Expected result has appeared
                        # Verify result doesn't appeared too late compared to accepted time
                        self.__raise_accepted_time_if_needed()
                        # Return obtained result
                        return result
                    
                if not stop_retry:
                    stop_retry = self.__manage_process_timeout(last_exception)
                    if stop_retry:
                        # Timeout is reached, but it is configured to not raise an exception ; return last polled result
                        return result
                    self.__wait_after_process(has_failure)
        finally:
            self.__end_time = Tools.timer_s()
            if self.run_counter > 1 or self.__interrupt_counter > 0:
                logger.info(f"redo [{self.name}] has made {self.run_counter} runs and {self.__interrupt_counter} interrupts (process timeout: {self.process_timeout} s)")
        
        return result

    def __execute_process(self):
        self.__current_process_result = None
        self.__current_process_exception = None
        
        try:
            # Execute before run
            self._execute_before_process()
            
            # Process
            self.__current_process_result = self._process()
            
            # Execute after run
            self._execute_after_process(self.__current_process_result)
        except Exception as exc:
            self.__current_process_exception = exc
    
    def __get_execute_thread(self):
        return FunctionThreaded(self.__execute_process, [])

    def __wait_process(self, thread, wait_interval_s = None):
        if wait_interval_s is None:
            refs = filter(lambda x: x is not None, (Config.redo_wait_process_max_interval_s, self.__interval/100, self.__interval_after_failure/100))
            wait_interval_s = max(Config.redo_wait_process_min_interval_s, min(refs))
            
        while True:
            if not thread.is_alive():
                break
            
            # Stop wait on timeout
            if Tools.timer_s() > self.start_time + self.timeout:
                logger.warning(f"redo [{self.name}] - run {self.run_counter} - stop waiting process, timeout of {self.timeout} s is reached")
                break
            
            # Stop wait on process timeout
            if self.spent_time > self.process_timeout:
                logger.warning(f"redo [{self.name}] - run {self.run_counter} - stop waiting process, process timeout of {self.process_timeout} s is reached")
                break
            
            time.sleep(wait_interval_s)
    
    def __interrupt_thread_if_still_alive(self, thread):
        res = False
        if thread.is_alive():
            self.__interrupt_counter += 1
            if thread.is_interruptable:
                thread.interrupt()
            
            # Manage specific interrupt
            # Note: for process implementation that launch uninterruptible code, this method can be used to stop the thread by calling thread.stop()
            self._process_interrupt(thread)
            
            res = True
            logger.warning(f"redo [{self.name}] interrupted after {self.spent_time} seconds ({self.run_counter} runs ; {self.__interrupt_counter} interrupts ; process timeout: {self.process_timeout} s)")
        return res
    
    def __post_process(self, interrupted):
        if interrupted:
            self.__current_process_exception = None
            
            # Manage exception logs
            self.__previous_process_exception_stack_trace = None
            self.__previous_process_exception_counter = 0
        elif self.__current_process_exception is None:
            if not self._is_redo_needed(self.__current_process_result):
#                            logger.trace("   --- redo [{}] -> return: {}", name, (self.__current_process_result != None ? self.__current_process_result.toString() : "None")))
                return True, self.__current_process_result
            
            # Manage exception logs
            self.__previous_process_exception_stack_trace = None
            self.__previous_process_exception_counter = 0
        else:
            # Manage exception logs
            current_process_exception_stack_trace = Tools.represent_exception(self.__current_process_exception)
            if self.__previous_process_exception_stack_trace is not None and current_process_exception_stack_trace == self.__previous_process_exception_stack_trace:
                self.__previous_process_exception_counter += 1
                
                if self.__previous_process_exception_counter % 10 == 0:
                    logger.warning(f"redo [{self.name}] - run {self.run_counter} - got previous process exception 10 more times ({self.__previous_process_exception_counter} times in total): {self.__get_error_description(self.__current_process_exception)}")
            else:
                logger.warning(f"redo [{self.name}] - run {self.run_counter} - got process exception {self.__get_error_description(self.__current_process_exception)}:\n{Tools.indent_string(4, current_process_exception_stack_trace)}")
                self.__previous_process_exception_stack_trace = current_process_exception_stack_trace
                self.__previous_process_exception_counter = 1
            
            raise self.__current_process_exception
        
        # Don't stop retries, and return current result
        return False, self.__current_process_result
    
    def __manage_process_exception(self, exc, last_exception):
        has_failure = False
        stop_retry = False
        
        # Begin by calling _execute_after_failure enabling to replace exception considered ignored or stopping retry
        try:
            self._execute_after_failure(exc)
        except Exception as exc_2:
            exc = exc_2
        
        # Manage stop, ignored, failure
        if self._is_stop_retry(exc):
            stop_retry = True
            self.__process_stop_retry(exc)
        elif self._is_ignored(exc):
            self.__process_ignored(exc)
        else:
            has_failure = True
            self.__process_failure(exc, last_exception)
        
        return has_failure, stop_retry
    
    def __process_ignored(self, exc):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"redo [{self.name}] - run {self.run_counter} - ignored process exception {self.__get_error_description(exc)}")
        self._execute_after_ignored(exc)
    
    def __process_stop_retry(self, exc):
        logger.info(f"redo [{self.name}] - run {self.run_counter} - stop retries due to exception {self.__get_error_description(exc)}")
        self._execute_after_stop_retry(exc)
        if self.__do_raise_on_stop_retry:
            raise exc
    
    def __process_failure(self, exception, last_exception):
        self.__failure_counter += 1
        if self.__failure_counter >= self.__nb_successive_failures_before_error:
            if self.__failure_counter > 1:
                logger.info(f"redo [{self.name}] - run {self.run_counter} - maximum successive failure ({self.__nb_successive_failures_before_error}) is reached - following previous failure is skipped: {self.__get_error_description(last_exception) if last_exception else '[None]'}")
            logger.info(f"redo [{self.name}] - run {self.run_counter} - maximum successive failure ({self.__nb_successive_failures_before_error}) is reached - following failure is considered as an error: {self.__get_error_description(exception)}")

            # Raise exception
            new_message = f"redo [{self.name}] - run {self.run_counter} - maximum successive failure ({self.__nb_successive_failures_before_error}) is reached"
            Tools.raise_same_exception_type(exception, new_message)
        else:
            if self.__failure_counter > 1:
                logger.info(f"redo [{self.name}] - run {self.run_counter} - {self.__failure_counter}'th successive failure - following previous failure is skipped: {self.__get_error_description(last_exception) if last_exception else '[None]'}")

    def __raise_accepted_time_if_needed(self):
        if self.accepted_time is not None and self.spent_time > self.accepted_time:
            accepted_msg = self._get_accepted_time_message()
            accepted_msg = f"{accepted_msg}\n -> " if accepted_msg is not None and len(accepted_msg) > 0 else ""
            msg = f"{accepted_msg}Too long after {self.spent_time} s in redo [{self.name}] ({self.run_counter} runs ; {self.__interrupt_counter} interrupts ; timeout: {self.timeout} s ; process timeout: {self.process_timeout} s ; allowed retries: {self.__nb_retries})"
            raise FunctionalException(msg)
    
    def __manage_process_timeout(self, last_exception):
        res = False
        if self.spent_time > self.timeout or self.__nb_retries > 0 and self.run_counter >= self.__nb_retries:
              if self.__func_to_call_on_timeout is not None:
                res = self.__func_to_call_on_timeout()
            elif self.__do_raise_on_timeout:
                timeout_msg = self._get_timeout_message()
                timeout_msg = f"{timeout_msg}\n -> " if timeout_msg is not None and len(timeout_msg) > 0 else ""
                msg = f"{timeout_msg}Timed out after {self.spent_time} s in redo [{self.name}] ({self.run_counter} runs ; {self.__interrupt_counter} interrupts ; timeout: {self.timeout} s ; process timeout: {self.process_timeout} s ; allowed retries: {self.__nb_retries})"
                raise TimeoutException(msg) from last_exception
            elif self.__do_raise_last_exception and last_exception is not None:
                raise last_exception
            else:
                # Return a timeout has occurred to stop retries
                res = True
        return res
    
    def __wait_after_process(self, has_failure):
        # Manage waits
        if has_failure:
            if self.__interval_after_failure > 0:
                try:
                    time.sleep(self.__interval_after_failure)
                except InterruptedError as e:
                    # Thread.currentThread().interrupt()
                    raise TechnicalException(self.__get_error_message(e)) from e
        elif self.__interval > 0:
            try:
                time.sleep(self.__interval)
            except InterruptedError as e:
                # Thread.currentThread().interrupt()
                raise TechnicalException(self.__get_error_message(e)) from e
    
    def _get_waited_description(self):
        return None
    
    def _get_timeout_message(self):
        waited_description = self._get_waited_description()
        if waited_description is not None:
            return f"Timeout ({self.timeout} s) when waiting {waited_description}"
        else:
            return None
    
    def _get_accepted_time_message(self):
        waited_description = self._get_waited_description()
        if waited_description is not None:
            return f"Too long ({self.spent_time} s) to wait {waited_description} (accepted time: {self.__accepted_time} s)"
        else:
            return None
    
    def __get_error_description(self, exc):
        return f"{Typing.get_object_class_name(exc)}('{self.__get_error_message(exc)}')"
        
    def __get_error_message(self, exc):
        if hasattr(exc, 'message'):
            return exc.message
        else:
            return str(exc)
        
    def _execute_before_process(self):
        """
        Method launched before each run.
        If an exception occurs in self method, the run will not be done.
        """
        # Nothing is done by default
        pass
        
    def _execute_after_process(self, result):
        """
        Method launched after each run.
        If an exception occurs in self method, it is managed as a run failure exception.
        """
        # Nothing is done by default
        pass
    
    def _execute_after_failure(self, exception):
        """
        Method launched after a failure and before considering if an exception is ignored or stopping retry.
        By implementing this method, it is possible to ignore/stop retry on an exception with a specific message,
        by replacing it by a custom exception (or dedicated RedoIgnoredException and RedoStopRetryException).
        """
        # Nothing is done by default
        pass
    
    def _execute_after_ignored(self, exception):
        """
        Method launched after an exception is ignored and before a redo.
        """
        # Nothing is done by default
        pass
    
    def _execute_after_stop_retry(self, exception):
        """
        Method launched after an exception is stopping retry.
        """
        # Nothing is done by default
        pass
    
    def with_unlimited_number_retries(self):
        """
        Sets unlimited number of allowed retries.
        @return A self reference.
        """
        return self.with_allowed_number_retries(0)
    
    def with_allowed_number_retries(self, nb_retries):
        """
        Sets the number of allowed retries.
        If not successful after self number of retries, a timeout exception is thrown.
        If the number of retries is set to 0, it retries infinitely until timeout.
        {@link #DEFAULT_NB_RETRIES}.
        @param nb_retries Number of allowed retries.
        @return A self reference.
        """
        self.__nb_retries = nb_retries
        return self
    
    def with_accepted_time(self, accepted_time):
        """
        Sets the max waiting time that is acceptable, otherwise it is considered as an error.
        @param accepted_time Accepted time in seconds.
        @return A self reference.
        """
        self.__accepted_time = accepted_time
        return self
    
    def with_timeout(self, timeout):
        """
        Sets how long to wait for the evaluated condition to be True. The default timeout is
        {@link #DEFAULT_TIMEOUT_S}.
        @param timeout Timeout in seconds.
        @return A self reference.
        """
        self.__timeout = timeout
        return self
    
    def update_timeout(self, timeout):
        """
        Update how long to still wait from now for the evaluated condition to be True.
        @param timeout Timeout in seconds.
        @return A self reference.
        """
        self.__timeout = self.spent_time + timeout
        return self
    
    def without_timeout(self):
        """
        @return A self reference.
        """
        self.__timeout = 24 * 3600
        return self

    def with_function_to_call_on_timeout(self, func=None):
        """ Define function to call on timeout.
        Note: if a function is defined, defined function is responsible to return a result or raise and exception
        @param func Function to call on timeout
        @return A self reference.
        """
        self.__func_to_call_on_timeout = func
        return self

    def with_raise_on_timeout(self, do_raise=True):
        """
        @param do_raise If raise the exception that has stopped retry
        @return A self reference.
        """
        self.__do_raise_on_timeout = do_raise
        return self

    def without_raise_on_timeout(self):
        """
        @return A self reference.
        """
        return self.with_raise_on_timeout(False)
    
    def with_process_in_thread(self, do_process_in_thread=True):
        """
        Set if 'process' must be executed in a thread.
        @return A self reference.
        """
        self.__do_process_in_thread = do_process_in_thread
        return self
    
    def without_process_in_thread(self):
        """
        Set 'process' must be executed without thread.
        @return A self reference.
        """
        return self.with_process_in_thread(do_process_in_thread=False)
    
    def with_process_timeout(self, timeout):
        """
        Sets how much time is let to process method until it is interrupted and relaunched. The default timeout is
        {@link #DEFAULT_PROCESS_TIMEOUT_S}.
        @param timeout Timeout in seconds.
        @return A self reference.
        """
        self.__process_timeout = timeout
        return self
    
    def with_process_timeout_for_retry(self, nb_retry):
        """
        Set appropriate process timeout in order to make at least given number of retries until global timeout.
        @param nb_retry Number of retries to do at least.
        @return A self reference.
        """
        return self.with_process_timeout(self.timeout / nb_retry)
    
    def polling_every(self, duration):
        """
        Sets how often the condition should be evaluated.
        
        In reality, the interval may be greater as the cost of actually evaluating a condition function
        is not factored in. The default polling interval is {@link #DEFAULT_POLLING_INTERVAL_S}.
        @param duration The duration between two polling.
        @return A self reference.
        """
        self.__interval = duration
        return self
    
    def with_nb_successive_failure_before_error(self, nb_successive_failures_before_error):
        """
        Sets the number of successive failures that must occur before considering being in error.
        {@link #DEFAULT_NB_SUCCESSIVE_FAILURES_BEFORE_ERROR}.
        @param nb_successive_failures_before_error Number of successive failures.
        @return A self reference.
        """
        self.__nb_successive_failures_before_error = nb_successive_failures_before_error
        return self
    
    def with_waiting_after_failure(self, duration):
        """
        Sets how much time must be waited after a failure before a new retry.
        The default interval is {@link #DEFAULT_WAITING_INTERVAL_AFTER_FAILURE_S}.
        @param duration Duration to wait in seconds.
        @return A self reference.
        """
        self.__interval_after_failure = duration
        return self
    
    def ignoring(self, *exception_types):
        """
        Configures to ignore specific types of exceptions while waiting for a condition.
        Any exceptions not whitelisted will be allowed to propagate, terminating the wait.
        @param exception_types The types of exceptions to ignore.
        @return A self reference.
        """
        if exception_types:
            self.__ignored_exceptions.extend(exception_types)
        return self
    
    def _is_ignored(self, exception):
        for exception_type in self.__ignored_exceptions:
            if isinstance(exception, exception_type):
                return True
        return False
    
    def stop_retry_on(self, *exception_types):
        """
        Configures to stop retries on specific types of exceptions while waiting for a condition.
        @param exception_types The types of exceptions that would stop retries.
        @return A self reference.
        """
        if exception_types:
            self.__stop_retry_exceptions.extend(exception_types)
        return self
    
    def _is_stop_retry(self, exception):
        for exception_type in self.__stop_retry_exceptions:
            if isinstance(exception, exception_type):
                return True
        return False
    
    def with_raise_on_stop_retry(self, do_raise=True):
        """
        @param do_raise If raise the exception that has stopped retry
        @return A self reference.
        """
        self.__do_raise_on_stop_retry = do_raise
        return self
    
    def without_raise_on_stop_retry(self):
        """
        @return A self reference.
        """
        return self.with_raise_on_stop_retry(False)
    
    def _print_run_counter(self):
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"   --- redo ({self.name}) : run {self.run_counter}")
    
        
        
        



"""
URL configuration for server_rest_example project.

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/5.2/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from django.contrib import admin
from django.urls import path
from rest_framework import routers
from django.urls.conf import include
from rest_api.application.apps import django_application_inst  # @UnresolvedImport
from holado_core.common.exceptions.technical_exception import TechnicalException

router = routers.DefaultRouter()
if django_application_inst is not None:
    django_application_inst.register_urls(router)
    print(f"router urls: {router.urls}")
else:
    raise TechnicalException("Django application instance is not set")

urlpatterns = [
    path('', include(router.urls)),
    path('api-auth/', include('rest_framework.urls', namespace='rest_framework')),
    path('admin/', admin.site.urls),
]



# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

# This file is a helper to build generic scripts using HolAdo solution.
# Usually, it is copied and adapted for specific usage.

import os
import logging
import argparse
from holado_helper.initialize_holado import insert_sys_path, insert_sys_paths


def _chdir(path):
    res = os.getcwd()
    os.chdir(path)
    return res

def _initialize_holado(TSessionContext=None, logging_config_file_path=None, log_level=logging.WARNING, log_in_file=False, with_session_path=False, **kwargs):
    import holado
    session_kwargs={'with_session_path':with_session_path or log_in_file}
    if kwargs:
        session_kwargs.update(kwargs)
    holado.initialize_for_script(TSessionContext=TSessionContext, logging_config_file_path=logging_config_file_path,
                                 log_level=log_level, log_in_file=log_in_file, log_on_console=not log_in_file, 
                                 session_kwargs=session_kwargs)
    
def initialize(work_dir_path=None, change_work_dir=True, 
               log_level=logging.INFO, log_in_file=False, 
               TSessionContext=None, additional_sys_paths=None, with_session_path=False):
    res = None
    sys_paths = []
    if work_dir_path:
        if change_work_dir:
            res = _chdir(work_dir_path)
        sys_paths = [work_dir_path]
    
    if additional_sys_paths:
        sys_paths.extend(additional_sys_paths)
    insert_sys_paths(sys_paths)
    
    logging_config_file_path=None
    for dir_path in [res, work_dir_path, os.getcwd()]:
        if dir_path is not None:
            file_path = os.path.join(dir_path, "logging.conf")
            if os.path.exists(file_path):
                logging_config_file_path = file_path
                break
    
    _initialize_holado(TSessionContext, logging_config_file_path, log_level, log_in_file, with_session_path)
    
    return res

def change_working_dir(work_dir_path=None):
    if work_dir_path:
        res = _chdir(work_dir_path)
        insert_sys_path(work_dir_path)
        return res
    else:
        return None

def parse_logging_args():
    log_parser = get_logging_argument_parser()
    res, _ = log_parser.parse_known_args()
    return res

def get_logging_argument_parser(description=None, additional_arguments=None, add_help=False):
    parser = argparse.ArgumentParser(description=description, formatter_class=argparse.RawTextHelpFormatter, add_help=add_help)

    parser.add_argument('-ll', '--log-level', choices=['ERROR', 'WARNING', 'INFO', 'DEBUG', 'TRACE'], dest='log_level', default='WARNING',
                   help='Log level')
    
    parser.add_argument('-lf', '--log-in-file', dest='log_in_file', default=False, action="store_true",
                   help='If specified, log in file rather than console')
    
    if additional_arguments:
        for add_arg in additional_arguments:
            parser.add_argument(*add_arg[0], **add_arg[1])
    
    return parser
    
def change_logging_config(log_level=logging.INFO, log_in_file=False):
    import holado
    holado.change_logging_config(log_level=log_level, log_in_file=log_in_file, log_on_console=not log_in_file)
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import abc
import logging
from datetime import timedelta, datetime
from typing import NamedTuple

from holado.common.handlers.undefined import default
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tools.tools import Tools
from holado_python.common.tools.datetime import DateTime


logger = logging.getLogger(__name__)

class TokenManager(object):
    """
    Manage tokens, by storing tokens and requesting/refreshing when needed.
    This class is abstract, method request_token needs to be overridden.
    Note: to allow the possibility to modify method signatures, all methods parameters are named.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self.__token_info_by_uid = {}

    def has_token(self, *, uid):
        return uid in self.__token_info_by_uid

    def get_token_info(self, *, uid, do_raise_exception=True):
        if self.has_token(uid=uid):
            res = self.__token_info_by_uid[uid]
        else:
            if do_raise_exception:
                raise TechnicalException(f"UID '{uid}' has no token")
            else:
                res = None
        if Tools.do_log(logger, logging.TRACE):
            logger.trace(f"Token info of UID '{uid}': {Tools.represent_object(res)}")
        return res

    def _set_token_info(self, *, uid, token, request_kwargs=None):
        self.__token_info_by_uid[uid] = self._new_token_info(token=token, request_kwargs=request_kwargs)

    def delete_token(self, *, uid):
        if self.has_token(uid=uid):
            del self.__token_info_by_uid[uid]
        else:
            raise TechnicalException(f"UID '{uid}' has no token")

    def get_token(self, *, uid, request_kwargs=None):
        """Get token for given UID, and if needed request token with given request_kwargs"""
        if self.has_token(uid=uid):
            return self.get_token_info(uid=uid).token
        else:
            request_kwargs = request_kwargs if request_kwargs is not None else {}
            res = self.request_token(**request_kwargs)
            self._set_token_info(uid=uid, token=res, request_kwargs=request_kwargs)
            return res

    def _new_token_info(self, *, token, request_kwargs=None):
        """Create a new token info object containing all needed information related to token.
        Note: Parameter 'request_kwargs' is not used by default, it exists for use cases needing to store some of these information.
              In this case, this method must be overridden with expected behavior.
        """
        res = NamedTuple("TokenInfo", token=dict, creation_date=datetime)
        res.token = token
        res.creation_date = DateTime.now()
        return res

    def request_token(self, **request_kwargs):
        raise NotImplementedError()


class ExpiringTokenManager(TokenManager):
    """
    Manage tokens with expiration.
    This class is abstract, methods request_token and request_refresh need to be overridden.
    """
    __metaclass__ = abc.ABCMeta

    GLOBAL_DEFAULT_NEEDED_TIME_S = 30

    def __init__(self, with_refresh=True):
        super().__init__()
        self.__default_needed_time_s = self.GLOBAL_DEFAULT_NEEDED_TIME_S
        self.__with_refresh = with_refresh

    @property
    def with_refresh(self):
        return self.__with_refresh

    @property
    def default_needed_time_s(self):
        return self.__default_needed_time_s

    @default_needed_time_s.setter
    def default_needed_time_s(self, needed_time_s):
        self.__default_needed_time_s= needed_time_s

    def get_token(self, *, uid, request_kwargs=None, needed_time_s=default):
        if self.has_token(uid=uid):
            if request_kwargs is not None:
                # Note: if request_kwargs is not defined, it is not possible to make any request
                self.ensure_token_is_valid(uid=uid, request_kwargs=request_kwargs, needed_time_s=needed_time_s)
            return self.get_token_info(uid=uid).token
        else:
            return super().get_token(uid=uid, request_kwargs=request_kwargs)

    def _new_token_info(self, *, token, request_kwargs=None):
        """Create a new token info object containing all needed information related to token.
        Note: This implementation prepare a NamedTuple with field 'expiration_date', and 'refresh_expiration_date' if with_refresh, but is not able to fill them,
              thus this method has to be overridden to complete these fields.
        """
        if self.__with_refresh:
            res = NamedTuple("TokenInfo", token=dict, creation_date=datetime, expiration_date=datetime, refresh_expiration_date=datetime)
        else:
            res = NamedTuple("TokenInfo", token=dict, creation_date=datetime, expiration_date=datetime)
        res.token = token
        res.creation_date = DateTime.now()
        return res

    def ensure_token_is_valid(self, *, uid, request_kwargs=None, needed_time_s=default):
        request_kwargs = request_kwargs if request_kwargs is not None else {}
        needed_time_s = needed_time_s if needed_time_s is not default else self.default_needed_time_s

        # Note: this method is supposed to be called when a token exists already, thus do_raise_exception is True
        ti = self.get_token_info(uid=uid, do_raise_exception=True)

        now = DateTime.now()
        needed_end = now + timedelta(seconds=needed_time_s)
        if needed_end > ti.expiration_date:
            if self.with_refresh and ti.refresh_expiration_date > now:
                token = self.request_refresh(uid=uid, **request_kwargs)
            else:
                token = self.request_token(**request_kwargs)
            self._set_token_info(uid=uid, token=token)

    def request_refresh(self, *, uid, **request_kwargs):
        """ Process refresh request.
        Note: given input kwargs are those used for the request.
        """
        raise NotImplementedError()






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_report.report.builders.report_builder import ReportBuilder
from holado_core.common.tools.tools import Tools
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_system.system.filesystem.file import File
from holado_test.common.context.context_tools import ContextTools

logger = logging.getLogger(__name__)



class ShortScenarioReportBuilder(ReportBuilder):
    def __init__(self, filepath, file_format='txt', exclude_statuses=None, exclude_categories=None, use_compact_format=True):
        self.__file_format = file_format.lower()
        self.__exclude_statuses = exclude_statuses
        self.__exclude_categories = exclude_categories
        self.__use_compact_format = use_compact_format
        
        if self.__file_format == 'txt':
            self.__file = File(filepath, mode='wt')
        else:
            raise TechnicalException(f"Unmanaged format '{self.__file_format}' (possible formats: 'txt')")
        
    def build(self):
        '''
        The file is built after each scenario
        '''
        pass
        
    def after_scenario(self, scenario, scenario_report=None):
        from holado_report.report.report_manager import ReportManager
        status_info = ReportManager.ScenarioTools.get_current_scenario_status_info(scenario)
        
        # Manage excluded scenarios
        if self.__exclude_statuses and status_info.validation_status in self.__exclude_statuses:
            return
        if status_info.validation_category is not None and self.__exclude_categories:
            ind = status_info.validation_category.find(' (')
            category = status_info.validation_category[:ind] if ind > 0 else status_info.validation_category
            if category in self.__exclude_categories:
                return
        
        self.__file_add_scenario(scenario, scenario_report, status_info)
        
    def after_all(self):
        # Manage file fail
        self.__file.close()
        
    def __file_add_scenario(self, scenario, scenario_report, status_info):
        from holado_report.report.report_manager import ReportManager
        
        self.__open_file_if_needed()
        
        msg_list = []
        category_str = f" => {status_info.validation_category}" if status_info.validation_category else ""
        if status_info.step_failed:
            msg_list.append(f"scenario in {ReportManager.ScenarioTools.format_scenario_short_description(scenario)} - {ReportManager.StepTools.format_step_short_description(status_info.step_failed, status_info.step_failed_nb, has_failed=True)} - {status_info.validation_status}{category_str}")
        else:
            msg_list.append(f"scenario in {ReportManager.ScenarioTools.format_scenario_short_description(scenario)} - {status_info.validation_status}{category_str}")
        msg_list.append(f"    Feature/Scenario: {scenario.feature.name}  =>  {scenario.name}")
        msg_list.append(f"    Report: {scenario_report.report_path}")
        msg_list.append(f"    Tags: -t " + " -t ".join(scenario.feature.tags + scenario.tags))
        
        if status_info.step_context and status_info.step_context.start_datetime is not None:
            msg_list.append(f"    Scenario/Step periods: {ContextTools.format_context_period(status_info.scenario_context, use_compact_format=self.__use_compact_format)} -> {ContextTools.format_context_period(status_info.step_context, dt_ref=status_info.scenario_context.start_datetime, use_compact_format=self.__use_compact_format)}")
        else:
            msg_list.append(f"    Scenario period: {ContextTools.format_context_period(status_info.scenario_context, use_compact_format=self.__use_compact_format)}")
        
        step_error_message = ReportManager.StepTools.get_step_error_message(status_info.step_failed)
        if step_error_message:
            if "\n" in step_error_message:
                msg_list.append(f"    Error message: ")
                msg_list.append(Tools.indent_string(8, step_error_message))
            else:
                msg_list.append(f"    Error message: {step_error_message}")
        
        if status_info.scenario_error is not None:
            if "\n" in status_info.scenario_error:
                msg_list.append(f"    Scenario error: ")
                msg_list.append(Tools.indent_string(8, status_info.scenario_error))
            else:
                msg_list.append(f"    Scenario error: {status_info.scenario_error}")
        
        msg_list.append(f"")
        msg_list.append(f"")
        
        self.__file.writelines(msg_list)
    
    def __open_file_if_needed(self):
        if not self.__file.is_open:
            self.__file.open()
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from builtins import super
from holado_core.common.tables.table_row import TableRow
from holado_core.common.tables.table import Table
from holado_core.common.exceptions.functional_exception import FunctionalException
import copy
from holado_core.common.exceptions.technical_exception import TechnicalException
import logging
from holado_core.common.tables.comparators.table_cell_comparator import TableCellComparator
from holado_core.common.tables.comparators.table_row_comparator import TableRowComparator
from holado_core.common.tools.tools import Tools
from holado_core.common.tables.comparators.table_with_header_comparator import TableWithHeaderComparator

logger = logging.getLogger(__name__)


class TableWithHeader(Table):
    
    def __init__(self):
        super().__init__()
        
        self.__header = TableRow()
        
    @property
    def header(self):
        return self.__header
    
    @header.setter
    def header(self, header):
        self.__header = header

    @Table.nb_columns.getter  # @UndefinedVariable
    def nb_columns(self):
        return len(self.header)
    
    def __copy__(self):
        res = super().__copy__()
        
        # Copy header
        res.header = copy.copy(self.header)
        
        return res
    
    def set_header(self, row=None, cells=None, cells_content=None):
        if row is not None:
            self.header = row
        else:
            self.header = TableRow(cells=cells, cells_content=cells_content)
    
    def add_row(self, row=None, cells=None, cells_content=None, contents_by_colname: dict =None, **kwargs):
        # Manage super call
        if contents_by_colname is None:
            return super().add_row(row=row, cells=cells, cells_content=cells_content, **kwargs)
        
        # Verify column names
        for colname in contents_by_colname:
            if not self.has_column(colname, raise_exception=False):
                raise FunctionalException(f"In parameter 'values_by_colname', the column name '{colname}' is not in table header")
        
        # Add row
        row_contents = [contents_by_colname[cn] if cn in contents_by_colname else None for cn in self.get_column_names()]
        return self.add_row(cells_content=row_contents, **kwargs)
        
    def extend(self, table, copy_method=None, **kwargs):
        # Verify table headers are identical
        comparator = TableWithHeaderComparator()
        comparator.equals_headers(table, self)
        
        super().extend(table, copy_method, **kwargs)
        
    def represent(self, indent = 0, limit_rows = -1, value_prefix = None, value_postfix = None):
        res_list = []
        
        res_list.append(self.header.represent(indent, value_prefix=value_prefix, value_postfix=value_postfix))
        res_list.append("\n")
        
        res_list.extend( super().represent(indent, limit_rows=limit_rows) )
        
        return "".join(res_list)
    
    def get_column(self, index=None, name=None, **kwargs):
        if name is not None:
            index = self.get_column_index(name)
        return super().get_column(index, **kwargs)
    
    def get_column_names(self):
        return [cell.string_content for cell in self.header]
    
    def get_column_indexes_by_string_content(self):
        return {cell.string_content : index for index, cell in enumerate(self.header)}
    
    def get_column_index(self, name):
        indexes = self.get_column_indexes_by_string_content()
        if name in indexes:
            return indexes[name]
        else:
            raise FunctionalException(f"Table has no column named '{name}' (existing column names: {self.header.represent()})")
    
    def has_column(self, name, raise_exception=True):
        try:
            self.get_column_index(name)
        except FunctionalException as exc:
            if raise_exception:
                raise exc
            else:
                return False
        else:
            return True

    def add_column(self, col=None, cells=None, cells_content=None, add_new_rows=False, name=None, **kwargs):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"add_column({col=}, {cells=}, {cells_content=}, {name=})")
        self.header.add_cell(content=name, **kwargs)
        super().add_column(col=col, cells=cells, cells_content=cells_content, add_new_rows=add_new_rows, **kwargs)

    def remove_column(self, index=None, name=None, **kwargs):
        if name is not None:
            index = self.get_column_index(name)
        self.header.remove_cell(index, **kwargs)
        return super().remove_column(index, **kwargs)
    
    def switch_columns(self, index_1=None, index_2=None, name_1=None, name_2=None, **kwargs):
        if name_1 is not None:
            index_1 = self.get_column_index(name_1)
        if name_2 is not None:
            index_2 = self.get_column_index(name_2)
        if index_1 == index_2:
            raise TechnicalException(f"The two indexes are equal (index = {index_1})")
    
        self.header.switch_cells(index_1, index_2, **kwargs)
        super().switch_columns(index_1, index_2, **kwargs)
    
    def sort(self, indexes=None, names=None, reverse=False, order_values_by_index=None, order_values_by_name=None, **kwargs):
        """Sort table rows according to indexes or column names order."""
        if names:
            indexes = tuple(self.get_column_index(name) for name in names)
        if order_values_by_name:
            order_values_by_index = {self.get_column_index(name):values for name,values in order_values_by_name.items()}
        super().sort(indexes=indexes, reverse=reverse, order_values_by_index=order_values_by_index, **kwargs)
    
    def is_sorted(self, indexes=None, names=None, reverse=False, **kwargs):  # @UnusedVariable
        """Check if table rows are sorted according to indexes or column names order."""
        if names:
            indexes = tuple(self.get_column_index(name) for name in names)
        return super().is_sorted(indexes=indexes, reverse=reverse, **kwargs)

    def order_columns(self, indexes=None, names=None, **kwargs):
        if names:
            indexes = tuple(self.get_column_index(name) for name in names)
        super().order_columns(indexes=indexes, **kwargs)
        
    def remove_rows_verifying(self, expected_table, table_comparator, keep_rows=False, **kwargs):
        """
        Remove rows verifying expected values.
        Parameter 'expected_table' has to refer all possible columns. 

        Remove rows that are equals compared to at least one line of expected_table.
        If keep_rows is True, keep the rows rather than remove them. 
        """

        # Verifying header are matching
        table_comparator.equals_headers(self, expected_table, raise_exception=True)

        # Remove rows
        super().remove_rows_verifying(expected_table, table_comparator, keep_rows=keep_rows, **kwargs)

    def remove_rows_only_verifying(self, expected_table, table_comparator, keep_rows=False, **kwargs):  # @UnusedVariable
        """
        Remove rows verifying expected values.
        Parameter 'expected_table' can contain only some columns. 

        Remove rows that are equals compared to at least one line of expected_table.
        If keep_rows is True, keep the rows rather than remove them. 
        """
        col_indexes = [self.get_column_index(hcell.string_content) for hcell in expected_table.header]
        n = 0
        while n < self.nb_rows:
            comp_row = copy.copy(self.get_row(n, **kwargs))
            # logger.info(f"+++++++++++++ line {n} - comp_row: {comp_row.represent(0)}")
            comp_row.order_cells(indexes=col_indexes)
            # logger.info(f"+++++++++++++ line {n} - comp_row: {comp_row.represent(0)}")
            comp_row.keep_cells(indexes=range(len(col_indexes)))
            # logger.info(f"+++++++++++++ line {n} - comp_row: {comp_row.represent(0)}")
            
            found = False
            for i in range(expected_table.nb_rows):
                found = table_comparator.row_comparator.equals(comp_row, expected_table.get_row(i), raise_exception=False, do_log=False)
                if found:
                    break
            
            # logger.info(f"+++++++++++++ line {n}: {found}")
            if found ^ keep_rows:
                self.remove_row(n)
            else:
                n += 1
                
    def remove_rows_duplicated(self, row_comparator=None, column_names=None, **kwargs):
        """
        Remove rows that are duplicates of previous rows.
        If row_comparator is None, column_names is used to create a row comparator that compares only given columns.
        """
        if row_comparator is None and column_names is None:
            raise TechnicalException("If parameter 'row_comparator' is not defined, at least column_names must be defined")
        
        if row_comparator is None and column_names is not None:
            cells_comparators = [TableCellComparator() if cn in column_names else None for cn in self.get_column_names()]
            row_comparator = TableRowComparator(cells_comparators=cells_comparators)
            
        super().remove_rows_duplicated(row_comparator=row_comparator, **kwargs)
    
    def extract_col(self, col_name, **kwargs):
        """
        Builds a new TableWithHeader from column col_name
        """
        if self.has_column(col_name, raise_exception=True):
            returnTable = TableWithHeader()
            returnTable.header = TableRow(cells_content=[col_name])
            
            col = self.get_column(name=col_name, **kwargs)
            for content in col.cells_content:
                returnTable.add_row(cells_content=content, **kwargs)
                
            return returnTable
    
    def rename_column(self, old_name, new_name, **kwargs):
        index=self.get_column_index(old_name)
        
        self.__header.get_cell(index).content = new_name
    
    
    
    


import os
import sys
import argparse
import logging

here = os.path.abspath(os.path.dirname(__file__))

holado_path = os.path.join(here, "..", "..")
sys.path.insert(0, os.path.join(holado_path, "src"))


from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_report.report.analyze.scenario_duration_manager import ScenarioDurationManager


def __main(args):
    if not os.path.exists(args.report_path):
        raise FunctionalException(f"Report path '{args.report_path}' doesn't exist")
    eh_path = os.path.join(args.report_path, "execution_historic.json")
    if not os.path.exists(eh_path):
        raise FunctionalException(f"Report path '{args.report_path}' doesn't contain an 'execution_historic.json' file")
    
    output_path = args.output_path
    if output_path is None:
        output_path = os.path.join(args.report_path, "scenario_duration_tags.csv")
        
    sdm = ScenarioDurationManager()
    sdm.import_execution_historic(eh_path)
    
    if args.all_scenarios:
        scenario_duration_tags = sdm.compute_scenario_duration_tags(
            duration_limit_tags = args.duration_tags, 
            default_tag = args.default_tag, 
            tag_prefix = args.tag_prefix, 
            missing_tag = True, 
            new_tag = True, 
            unchanged_tag = True, 
            with_failed = True,
            add_apply = args.add_apply_column)
    else:
        scenario_duration_tags = sdm.compute_scenario_duration_tags(
            duration_limit_tags = args.duration_tags, 
            default_tag = args.default_tag, 
            tag_prefix = args.tag_prefix,
            add_apply = args.add_apply_column)
    
    sdm.create_file_scenario_duration_tags(output_path, scenario_duration_tags)
    
    logging.info(f"Finished")
    

if __name__ == "__main__":
    # Arguments
    descr = """Generate file with scenario duration tags.
    It needs as input the folder of the test report for which generating scenario duration tags.

    Example 1: Generate file with duration tags [(0.2, 'very_fast'), (1, 'fast'), (10, 'normal'), (60, 'slow')] and default tag 'very_slow' 
               
            python generate_scenario_duration_tags.py -r {path_to_report} -t 0.2:very_fast 1:fast 10:normal 60:slow -d very_slow
            
    Example 2: Same as Example 1 but with specific output filename. 
    
            python generate_scenario_duration_tags.py -r {path_to_report} -t 0.2:very_fast 1:fast 10:normal 60:slow -d very_slow -o {path_to_report}/scenario_duration_tags.csv
            
        Note: With this "-o" parameter, it is equivalent to Example 1.
    """
    
            
    parser = argparse.ArgumentParser(description=descr, formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-r', '--report', dest='report_path', required=True,
                   help='Path to the test report folder.')
    
    parser.add_argument('-o', '--output', dest='output_path', default=None,
                   help='Path to the output file. If not defined, a new file will be generated in report folder')

    def duration_tag(s):
        try:
            dur, tag = s.split(':')
            return float(dur), tag
        except:
            raise argparse.ArgumentTypeError("Duration tag must in format '{int}:{str}'")
    parser.add_argument('-t', '--duration-tags', dest='duration_tags', type=duration_tag, nargs="*", required=True,
                   help='List of tuples (duration, tags). If a scenario has a duration inferior to tuple duration, tuple tag is associated to the scenario.')
    parser.add_argument('-d', '--default-tag', dest='default_tag', required=True, 
                   help="Tag associated to scenario whose duration is greater than all durations of parameter '--duration-tags'.")

    parser.add_argument('-a', '--add-apply-column', dest='add_apply_column', default=False, action="store_true",
                   help="If set, the column 'Apply' will be added in output file.")
    
    parser.add_argument('-tp', '--tag-prefix', dest='tag_prefix', default="ScenarioDuration=", 
                   help="Prefix of the duration tag in .feature files.")
    parser.add_argument('-as', '--all-scenarios', dest='all_scenarios', default=False, action="store_true",
                   help='If set, all scenario are in output file, otherwise output file contains only scenarios that have passed and changing tag.')

    args = parser.parse_args()

    logging.basicConfig(format = '{asctime:s} | {thread:-5d} | {levelname:5s} | {module:35s} | {message:s}', style = '{', level=logging.DEBUG, handlers=[logging.StreamHandler()])
    
    # Execute
    try:
        __main(args)
    except:
        logging.exception("An error occured")
        sys.exit(1)

    sys.exit(0)
    
    


# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


from holado.common.context.session_context import SessionContext
from holado_test.behave.behave import *  # @UnusedWildImport
import logging
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_value.common.tables.converters.value_table_converter import ValueTableConverter
from holado_crypto.crypto.key import KeyManager
from holado_test.scenario.step_tools import StepTools

logger = logging.getLogger(__name__)


def __get_scenario_context():
    return SessionContext.instance().get_scenario_context()

def __get_text_interpreter():
    return __get_scenario_context().get_text_interpreter()

def __get_variable_manager():
    return __get_scenario_context().get_variable_manager()



@Given(r"generate new self-signed key files for localhost")
def step_impl(context):
    """Generate public and private keys for a localhost use.
    """
    table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
    kwargs = ValueTableConverter.convert_name_value_table_2_dict(table)

    KeyManager.generate_new_self_signed_key_files_for_localhost(**kwargs)

@Given(r"(?P<var_name>{Variable}) = public key for (?:(?P<is_x509_str>X509) )?PEM file (?P<file_path>{Str})")
def step_impl(context, var_name, is_x509_str, file_path):
    """Get public key instance.
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    is_x509 = is_x509_str is not None
    file_path = StepTools.evaluate_scenario_parameter(file_path)
    
    res = KeyManager.load_pem_public_key(file_path=file_path, is_x509=is_x509)
    
    __get_variable_manager().register_variable(var_name, res)

@Given(r"(?P<var_name>{Variable}) = private key for PEM file (?P<file_path>{Str})")
def step_impl(context, var_name, file_path):
    """Get private key instance.
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    file_path = StepTools.evaluate_scenario_parameter(file_path)
    
    res = KeyManager.load_pem_private_key(file_path=file_path)
    
    __get_variable_manager().register_variable(var_name, res)

@Given(r"(?P<var_name>{Variable}) = convert public key (?P<pub_key>{Variable}) to hexadecimal")
def step_impl(context, var_name, pub_key):
    """Return path to default cert file.
    Note: It uses ssl.get_default_verify_paths method.
    """
    var_name = StepTools.evaluate_variable_name(var_name)
    key = StepTools.evaluate_variable_value(pub_key)
    
    res = KeyManager.convert_public_key_to_hex(key)
    
    __get_variable_manager().register_variable(var_name, res)





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import copy
from holado_core.common.finders.tools.find_updater import FindUpdater


class FindParameters(object):
    """ Find parameters
    """
    
    _do_analyze_time_finder = True
    _do_analyze_time_criteria = True

    __instance_default = None
    __instance_default_without_raise = None
    
    def __init__(self, raise_no_such_element=True, raise_intermediate_no_such_element=True, find_level=0, visibility=True, analyze_time_spent=True, 
                 redo=False, redo_ignored_exceptions=None, nb_max_elements=None, nth_element=None, add_valid_container=False):
        """
        Constructor
        @param raise_no_such_element If True, raise NoSuchElementException, else return None, when no element is found.
        @param raise_intermediate_no_such_element If True, raise NoSuchElementException, else return None, when no element is found in intermediate searches.
        @param find_level Find level, beginning at 0 and increasing at each sub-find.
        @param visibility Visibility of elements to find (None means any visibility).
        @param analyze_time_spent If True, analyze time spent in find.
        @param redo If redo
        @param redo_ignored_exceptions Ignorded exception in redo 
        @param nb_max_elements Number max of elements to find
        @param nth_element N'th element to find
        @param add_valid_container If True, add container to find result if it validates criteria
        """
        self.raise_no_such_element = raise_no_such_element
        self.raise_intermediate_no_such_element = raise_intermediate_no_such_element
        self.find_level = find_level
        self.visibility = visibility
        self.analyze_time_spent = analyze_time_spent
        self.redo = redo
        self.redo_ignored_exceptions = []
        if redo_ignored_exceptions is not None:
            self.redo_ignored_exceptions.extend(redo_ignored_exceptions)
        self.nb_max_elements = nb_max_elements
        self.nth_element = nth_element
        self.add_valid_container = add_valid_container
    
    def update(self, updater:FindUpdater):
        """
        @param updater Find updater
        @return Updated parameters
        """
        res = self
        if updater.has_context_value("analyze_time_spent"):
            res = res.with_analyze_time_spent(updater.get_context_value("analyze_time_spent"))
        if updater.has_context_value("nb_max_elements"):
            res = res.with_nb_max_elements(updater.get_context_value("nb_max_elements"))
        if updater.has_context_value("nth_element"):
            res = res.with_nth_element(updater.get_context_value("nth_element"))
        if updater.has_context_value("redo"):
            res = res.with_redo(updater.get_context_value("redo"))
        if updater.has_context_value("ignoring"):
            res = res.ignoring(updater.get_context_value("ignoring"))
        if updater.has_context_value("raise"):
            res = res.with_raise(updater.get_context_value("raise"))
        if updater.has_context_value("visibility"):
            res = res.with_visibility(updater.get_context_value("visibility"))
        return res

    def get_next_level(self, raise_exception=None):
        """
        @param raise_exception Whether raise exceptions or not
        @return Find parameter for next find level
        """
        res = copy.deepcopy(self)
        res.find_level += 1
        res.analyze_time_spent = False
        if raise_exception is not None:
            res.raise_no_such_element = raise_exception
            res.raise_intermediate_no_such_element = raise_exception
        return res
    
    def get_criteria_parameters(self):
        """
        @return Associated criteria parameters
        """
        res = copy.deepcopy(self)
        res.find_level += 1
        return res
    
    def with_visibility(self, visibility):
        """
        Note: if wanted visibility is the same, self instance is returned, else a new one is returned
        @param visibility Wanted visibility
        @return Same parameters but with given visibility
        """
        if visibility == self.visibility:
            return self
        else:
            res = copy.deepcopy(self)
            res.visibility = visibility
            return res
        
    def with_analyze_time_spent(self, analyze_time_spent):
        """
        Note: if wanted analyze_time_spent is the same, self instance is returned, else a new one is returned
        @param analyze_time_spent If analyze time spent
        @return Same parameters but with given analyze_time_spent
        """
        if analyze_time_spent == self.analyze_time_spent:
            return self
        else:
            res = copy.deepcopy(self)
            res.analyze_time_spent = analyze_time_spent
            return res
        
    def with_raise(self, raise_exception=True):
        """
        Note: if raise booleans are same, self instance is returned, else a new one is returned
        @param raise_exception Whether raise exceptions or not
        @return Same parameters but with raise
        """
        if self.raise_no_such_element == raise_exception and self.raise_intermediate_no_such_element == raise_exception:
            return self
        else:
            res = copy.deepcopy(self)
            res.raise_no_such_element = raise_exception
            res.raise_intermediate_no_such_element = raise_exception
            return res
        
    def without_raise(self):
        """
        Note: if raise booleans are already False, self instance is returned, else a new one is returned
        @return Same parameters but without raise
        """
        return self.with_raise(False)
    
    def with_redo(self, redo):
        """
        Note: if wanted redo is the same, self instance is returned, else a new one is returned
        @param redo If do redo
        @return Same parameters but with given redo
        """
        if redo == self.redo:
            return self
        else:
            res = copy.deepcopy(self)
            res.redo = redo
            res.redo_ignored_exceptions = self.redo_ignored_exceptions if redo else None
            return res
        
    def ignoring(self, exception_class):
        """
        @param exceptionType Type of exception to ignore.
        @return Same parameters but with given ignoring
        """
        if exception_class in self.redo_ignored_exceptions:
            return self
        else:
            res = copy.deepcopy(self)
            res.redo_ignored_exceptions.append(exception_class)
            return res
        
    def with_nb_max_elements(self, nb_max_elements):
        """
        Note: if wanted number max of elements is the same, self instance is returned, else a new one is returned
        @param nb_max_elements Number max of elements to find
        @return Same parameters but with given number max of elements
        """
        if nb_max_elements == self.nb_max_elements:
            return self
        else:
            res = copy.deepcopy(self)
            res.nb_max_elements = nb_max_elements
            return res
        
    def with_nth_element(self, nth_element):
        """
        Note: if wanted n'th element is the same, self instance is returned, else a new one is returned
        @param nth_element N'th element to find
        @return Same parameters but with given number max of elements
        """
        if nth_element == self.nth_element:
            return self
        else:
            res = copy.deepcopy(self)
            res.nth_element = nth_element
            return res
        
    def with_valid_container(self, add_valid_container):
        """
        @param add_valid_container N'th element to find
        @return Same parameters but with or without container validating criteria
        """
        if add_valid_container == self.add_valid_container:
            return self
        else:
            res = copy.deepcopy(self)
            res.add_valid_container = add_valid_container
            return res
        
    @staticmethod
    def default(raise_exception=True):
        """
        @param raise_exception Whether raise exceptions or not
        @return Default find parameters with given raiseability
        """
        if raise_exception:
            return FindParameters.default_with_raise()
        else:
            return FindParameters.default_without_raise()
    
    @staticmethod
    def default_with_raise():
        """
        @return Default find parameters with raises
        """
        if FindParameters.__instance_default is None:
            FindParameters.__instance_default = FindParameters(True, True, 0, True, FindParameters._do_analyze_time_finder, False, None, None, None, False)
        return FindParameters.__instance_default
    
    @staticmethod
    def default_without_raise():
        """
        @return Default find parameters without any raise
        """
        if FindParameters.__instance_default_without_raise is None:
            FindParameters.__instance_default_without_raise = FindParameters(False, False, 0, True, FindParameters._do_analyze_time_finder, False, None, None, None, False)
        return FindParameters.__instance_default_without_raise
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado.common.handlers.enums import AutoNumber
import platform
from holado_core.common.exceptions.technical_exception import TechnicalException
import time
import os
from holado_core.common.tools.tools import Tools
import threading
from holado_core.common.tools.converters.converter import Converter

logger = logging.getLogger(__name__)



class OSTypes(AutoNumber):
    """
    Types of operating systems.
    """
    Windows = ()
    MacOS = ()
    Linux = ()
    Other = ()


class GlobalSystem(object):
    """
    System methods.
    """
    __detected_os_type:OSTypes = None
    __detected_os_version = None

    @classmethod
    def get_os_type(cls):
        """
        Detect the operating system from the os.name System property and cache
        the result
        @return the operating system detected
        """
        if cls.__detected_os_type is None:
            os_ = platform.system()
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"Detected OS: {os_}")
            if os_ == 'Darwin':
                cls.__detected_os_type = OSTypes.MacOS
            elif os_ == 'Windows':
                cls.__detected_os_type = OSTypes.Windows
            elif os_ == 'Linux':
                cls.__detected_os_type = OSTypes.Linux
            else:
                cls.__detected_os_type = OSTypes.Other
            logger.info(f"Detected OS type: {cls.__detected_os_type.name}")
        return cls.__detected_os_type

    @classmethod
    def get_os_version(cls):
        """
        Detect the operating system from the os.name System property and cache
        the result
        @return the operating system detected
        """
        if cls.__detected_os_version is None:
            cls.__detected_os_version = platform.release()
            if '-' in cls.__detected_os_version:
                cls.__detected_os_version = cls.__detected_os_version[:cls.__detected_os_version.index('-')]
            logger.info(f"Detected OS version: {cls.__detected_os_version}")
        return cls.__detected_os_version

    @classmethod
    def execute_curl(cls, curl_parameters, do_log_output=False):
        cmd = f"curl {curl_parameters}"
        return cls.execute_command(cmd, do_log_output=do_log_output, do_raise_on_stderr=False)

    @classmethod
    def execute_command(cls, cmd, do_log_output=False, do_raise_on_stderr=False):
        from holado_system.system.command.command import Command, CommandStates
        from holado_system.system.command.command_result import CommandResult
        
        command = Command(cmd, do_log_output=do_log_output, do_raise_on_stderr=do_raise_on_stderr)
        command.start()
        command.join()
        
        if command.state == CommandStates.Error:
            raise TechnicalException(f"Error while executing command [{cmd}]: [{command.error if command.error else command.stderr}]")
        return CommandResult(cmd, command.stdout, command.stderr)
    
    @classmethod
    def log_resource_usage(cls, prefix=None, level=logging.DEBUG, logger_=None):
        # cls.log_memory_usage(prefix, level, logger_)
        
        # Log all resources usage of this process and children processes
        if cls.get_os_type() == OSTypes.Linux:
            import resource
            cls.__log_resource_usage(resource.RUSAGE_SELF, None, None, prefix, level, logger_)
            cls.__log_resource_usage(resource.RUSAGE_CHILDREN, None, None, prefix, level, logger_)
        else:
            raise NotImplementedError(f"Not implemented for OS type '{cls.get_os_type().name}'")
        
        # Log threads
        cls.__log_threads(prefix, level, logger_)
    
    @classmethod
    def log_memory_usage(cls, prefix=None, level=logging.DEBUG, logger_=None):
        if cls.get_os_type() == OSTypes.Linux:
            import resource
            cls.__log_resource_usage(resource.RUSAGE_SELF, 'maxrss', 'maximum resident set size', prefix, level, logger_)
            cls.__log_resource_usage(resource.RUSAGE_CHILDREN, 'maxrss', 'maximum resident set size', prefix, level, logger_)
        else:
            raise NotImplementedError(f"Not implemented for OS type '{cls.get_os_type().name}'")
    
    @classmethod
    def __log_resource_usage(cls, which_process=None, resource_attr_name=None, resource_description=None, prefix=None, level=logging.DEBUG, logger_=None):
        if logger_ is None:
            logger_ = logger
        if not logger_.isEnabledFor(level):
            return
        
        if cls.get_os_type() == OSTypes.Linux:
            import resource
            if which_process is None:
                which_process = resource.RUSAGE_SELF
            
            ru = resource.getrusage(which_process)
            field_names = ['utime', 'stime', 'maxrss', 'minflt', 'majflt', 'inblock', 'oublock', 'nvcsw', 'nivcsw']
            resource_usage = {fn:getattr(ru, f'ru_{fn}') for fn in field_names}
            
            if which_process == resource.RUSAGE_SELF:
                usage_descr = "Process resource usage"
            elif which_process == resource.RUSAGE_CHILDREN:
                usage_descr = "Child processes resource usage"
            else:
                usage_descr = "Other resource usage"
            if resource_attr_name is not None:
                logger_.log(level, f"{prefix if prefix else ''}{usage_descr}: {resource_description} = {getattr(resource_usage, resource_attr_name)}")
            else:
                logger_.log(level, f"{prefix if prefix else ''}{usage_descr}: {resource_usage}")
        else:
            raise NotImplementedError(f"Not implemented for OS type '{cls.get_os_type().name}'")
    
    @classmethod
    def __log_threads(cls, prefix=None, level=logging.DEBUG, logger_=None):
        if logger_ is None:
            logger_ = logger
        if not logger_.isEnabledFor(level):
            return
        
        thread_names = [t.name for t in threading.enumerate()]
        logger_.log(level, f"{prefix if prefix else ''}Process has {len(thread_names)} threads: {thread_names}")
        
    @classmethod
    def yield_processor(cls):
        if cls.get_os_type() == OSTypes.Windows:
            time.sleep(0.0001)
        else:
            os.sched_yield()
    
    @classmethod
    def get_used_ports(cls):
        if cls.get_os_type() == OSTypes.Linux:
            cmd = "netstat -tunlep | grep LISTEN | awk '{print $4}' | awk '{gsub(\".*:\",\"\");print}'"
            result = cls.execute_command(cmd, do_log_output=False, do_raise_on_stderr=False)
            res = {Converter.to_integer(v) for v in result.output.split('\n') if len(v) > 0}
            return res
        else:
            raise NotImplementedError(f"Not implemented for OS type '{cls.get_os_type().name}'")

    @classmethod
    def get_first_available_anonymous_port(cls):
        return cls.get_available_anonymous_port(index=0)

    @classmethod
    def get_available_anonymous_port(cls, index=0):
        return cls.get_available_port(ports=range(32768, 65536), index=index)

    @classmethod
    def get_available_port(cls, ports=range(32768, 65536), index=0):
        used_ports = cls.get_used_ports()
        for port in ports:
            if port not in used_ports:
                if index == 0:
                    return port
                else:
                    index -= 1
        return None




# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import csv
from holado_core.common.tables.table_with_header import TableWithHeader
from holado_core.common.tables.table_row import TableRow
from holado.common.context.session_context import SessionContext
from holado_core.common.tables.table import Table
from holado_core.common.tables.table_manager import TableManager
from holado_data.data.generator.base import BaseGenerator

logger = logging.getLogger(__name__)




class CsvManager():
    # Default settings for CSV files: dialect Excel but with ';' delimiter
    default_dialect = "excel"
    default_delimiter = ";"
    default_quotechar = None
    default_quoting = None

    @classmethod
    def __get_path_manager(cls):
        return SessionContext.instance().path_manager
    
    @classmethod
    def _get_csv_kwargs(cls, dialect, delimiter, quotechar, quoting):
        csv_kwargs = {}
        
        if dialect is not None:
            csv_kwargs['dialect'] = dialect
        elif cls.default_dialect is not None:
            csv_kwargs['dialect'] = cls.default_dialect
            
        if delimiter is not None:
            csv_kwargs['delimiter'] = delimiter
        elif cls.default_delimiter is not None:
            csv_kwargs['delimiter'] = cls.default_delimiter
            
        if quotechar is not None:
            csv_kwargs['quotechar'] = quotechar
        elif cls.default_quotechar is not None:
            csv_kwargs['quotechar'] = cls.default_quotechar
            
        if quoting is not None:
            csv_kwargs['quoting'] = quoting
        elif cls.default_quoting is not None:
            csv_kwargs['quoting'] = cls.default_quoting
            
        return csv_kwargs
        
    @classmethod
    def table_with_content_of_CSV_file(cls, path, encoding=None, dialect=None, delimiter=None, quotechar=None, quoting=None, with_header=True):
        csv_kwargs = cls._get_csv_kwargs(dialect, delimiter, quotechar, quoting)
        
        with open(path, 'r', encoding=encoding) as fin:
            if with_header:
                dr = csv.DictReader(fin, **csv_kwargs)
                
                res = TableWithHeader()
                res.header = TableRow(cells_content=dr.fieldnames)
                
                for row in dr:
                    nrow=dict(row)
                    res.add_row(contents_by_colname=nrow)
            else:
                reader = csv.reader(fin, **csv_kwargs)
                
                res = Table()
                for row in reader:
                    res.add_row(cells_content=row)
        
        return res
        
    @classmethod
    def dict_list_with_content_of_CSV_file(cls, path, encoding=None, dialect=None, delimiter=None, quotechar=None, quoting=None):
        dg = cls.dict_generator_with_content_of_CSV_file(path, encoding, dialect, delimiter, quotechar, quoting)
        return [d for d in dg]
    
    @classmethod
    def dict_generator_with_content_of_CSV_file(cls, path, encoding=None, dialect=None, delimiter=None, quotechar=None, quoting=None):
        csv_kwargs = cls._get_csv_kwargs(dialect, delimiter, quotechar, quoting)
        
        class DictCSVGenerator(BaseGenerator):
            def __init__(self, path, encoding, csv_kwargs):
                super().__init__(name="dict CSV generator")
                
                self.__path = path
                self.__encoding = encoding
                self.__csv_kwargs = csv_kwargs
                
                self.__fin = None
                self.__dict_reader = None
            
            def __open_file_if_needed(self):
                if self.__fin is None:
                    self.__fin = open(self.__path, 'r', encoding=self.__encoding)
                    self.__dict_reader = csv.DictReader(self.__fin, **self.__csv_kwargs)
            
            def __close_file(self):
                self.__dict_reader = None
                self.__fin.close()
                self.__fin = None
                
            def __next__(self):
                self.__open_file_if_needed()
                try:
                    return dict(next(self.__dict_reader))
                except StopIteration as exc:
                    self.__close_file()
                    raise exc
        
        return DictCSVGenerator(path, encoding, csv_kwargs)
    
    @classmethod
    def tuple_list_with_content_of_CSV_file(cls, path, encoding=None, dialect=None, delimiter=None, quotechar=None, quoting=None):
        tg = cls.tuple_generator_with_content_of_CSV_file(path, encoding, dialect, delimiter, quotechar, quoting)
        return [t for t in tg]
    
    @classmethod
    def tuple_generator_with_content_of_CSV_file(cls, path, encoding=None, dialect=None, delimiter=None, quotechar=None, quoting=None):
        csv_kwargs = cls._get_csv_kwargs(dialect, delimiter, quotechar, quoting)
        
        class TupleCSVGenerator(BaseGenerator):
            def __init__(self, path, encoding, csv_kwargs):
                super().__init__(name="tuple CSV generator")
                
                self.__path = path
                self.__encoding = encoding
                self.__csv_kwargs = csv_kwargs
                
                self.__fin = None
                self.__reader = None
            
            def __open_file_if_needed(self):
                if self.__fin is None:
                    self.__fin = open(self.__path, 'r', encoding=self.__encoding)
                    self.__reader = csv.reader(self.__fin, **self.__csv_kwargs)
            
            def __close_file(self):
                self.__reader = None
                self.__fin.close()
                self.__fin = None
                
            def __next__(self):
                self.__open_file_if_needed()
                try:
                    return tuple(next(self.__reader))
                except StopIteration as exc:
                    self.__close_file()
                    raise exc
        
        return TupleCSVGenerator(path, encoding, csv_kwargs)
    
    @classmethod
    def create_csv_file(cls, path, table, encoding=None, dialect=None, delimiter=None, quotechar=None, quoting=None):
        csv_kwargs = cls._get_csv_kwargs(dialect, delimiter, quotechar, quoting)
        
        cls.__get_path_manager().makedirs(path)
        
        with open(path, 'w', encoding=encoding) as fout:
            if TableManager.is_table_with_header(table):
                fieldnames = table.header.cells_content
                dw = csv.DictWriter(fout, fieldnames=fieldnames, **csv_kwargs)
                
                dw.writeheader()
                for row in table.rows:
                    rowdict = {fieldnames[i]:row[i].string_content for i in range(len(fieldnames))}
                    dw.writerow(rowdict)
            else:
                writer = csv.writer(fout, **csv_kwargs)
                for row in table.rows:
                    cells_content = [c.string_content for c in row]
                    writer.writerow(cells_content)
    
    @classmethod
    def merge_csv(cls, input_paths, output_path=None, sort_column=None, with_header=True):
        input_tables = [cls.table_with_content_of_CSV_file(ip, with_header=with_header) for ip in input_paths]
        
        # Merge tables
        output_table = input_tables[0]
        for i in range(1, len(input_tables)):
            output_table.extend(input_tables[i])
        
        # Sort rows
        if sort_column is not None:
            output_table.sort(names=[sort_column])
        
        if output_path is not None:
            cls.create_csv_file(output_path, output_table)
        else:
            return output_table
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import abc
from holado_ui.ui.drivers.ui_internal_api import UIInternalAPI
from holado_core.common.finders.tools.find_parameters import FindParameters
import os
from holado_python.common.tools.datetime import DateTime

logger = logging.getLogger(__name__)



class GUIInternalAPI(UIInternalAPI):
    """ Base class for GUI internal API.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, driver):
        super().__init__(driver)
        
        self.__activate_wait_until_window_is_loaded = True
    
    @property
    def activate_wait_until_window_is_loaded(self):
        return self.__activate_wait_until_window_is_loaded
    
    @activate_wait_until_window_is_loaded.setter
    def activate_wait_until_window_is_loaded(self, activate_wait_until_window_is_loaded):
        """
        @param activate_wait_until_window_is_loaded True/false for respectively activation/deactivation of wait until window is loaded
        """
        self.__activate_wait_until_window_is_loaded = activate_wait_until_window_is_loaded
    
    def get_element_coordinates(self, element):
        """
        @param element Element
        @return Element coordinates
        """
        raise NotImplementedError

    def get_element_description(self, element, with_text=True):
        """
        @param element Element
        @param with_text If true add the text of the element
        @return Element description
        """
        raise NotImplementedError
    
    def get_element_location(self, element):
        """
        @param element Element
        @return Element location
        """
        raise NotImplementedError
    
    def get_element_rectangle(self, element):
        """
        @param element Element
        @return Element rectangle
        """
        raise NotImplementedError
    
    def get_element_size(self, element):
        """
        @param element Element
        @return Element size
        """
        raise NotImplementedError

    def get_element_text(self, element):
        """
        @param element Element
        @return Text of given element
        """
        find_context = self.driver.find_builder.context()
        find_parameters = FindParameters.default_with_raise().with_visibility(True)
        return self._get_element_text(element, find_context, find_parameters)

    def _get_element_text(self, element, find_context, find_parameters):
        """
        @param element Element
        @param find_context Find context
        @param find_parameters Find parameters
        @return Text of given element
        """
        raise NotImplementedError
    
    def clear_element(self, element):
        """
        Clear content of editable element
        @param element Element
        """
        raise NotImplementedError
    
    def make_screenshot(self, destination_path, context_description):
        """
        Make a screenshot of current window.
        @param destination_path Destination path
        @param context_description Context description that will be inserted in file names
        """
        # Create file name
        date_str = DateTime.now().strftime('%Y-%m-%d_%H-%M-%S.%f')
        file_name = f"{date_str}-{context_description}-screenshot.png"
        screenshot_path = os.path.join(destination_path, file_name)
        
        # Take screenshot
        self.get_screenshot_file(screenshot_path)
    
    def get_screenshot_file(self, screenshot_path, element=None, margin_around_element=0):
        """
        Make a screenshot and return image file instance.
        @param screenshot_path Screenshot path
        @param element Element to screenshot (if not defined, a screenshot of whole page is done)
        @param margin_around_element Size of the marge around element to include in screenshot  
        @return File containing screenshot
        """
        raise NotImplementedError
    
    def get_screenshot_file_in_destination(self, destination_path, element=None, margin_around_element=0):
        """
        Make a screenshot of given element and return image file instance.
        @param destination_path Destination path
        @param element Element to screenshot (if not defined, a screenshot of whole page is done)
        @param margin_around_element Size of the marge around element to include in screenshot  
        @return File containing screenshot
        """
        # Format now datetime
        date_str = DateTime.now().strftime('%Y-%m-%d_%H-%M-%S.%f')
        if element is not None:
            file_name = f"{date_str}-element-margin{margin_around_element}.png"
        else:
            file_name = f"{date_str}-page.png"
        screenshot_path = os.path.join(destination_path, file_name)
        
        return self.get_screenshot_file(screenshot_path, element, margin_around_element)
    
    def is_displayed(self, element):
        """
        @param element An element.
        @return True if given element is displayed.
        """
        raise NotImplementedError
    
    def is_visible(self, element, find_context, find_parameters):
        """
        Return if element visibility corresponds to the one defined in find parameters.
        Note: If find_parameters.visibility is null, always return true
        @param element An element.
        @param find_context Find context
        @param find_parameters Find parameters
        @return True if given element is visible.
        """
        if find_parameters.visibility is None:
            return True
        
        return (find_parameters.visibility == self.is_displayed(element))
    
    def move_mouse_over(self, element):
        """
        Move mouse over a given element
        @param element Element 
        """
        raise NotImplementedError

    def scroll_by(self, x, y, sleep_after_scroll_ms=100):
        """
        Scroll by given number of pixels
        @param x X number of pixels
        @param y Y number of pixels
        @param sleep_after_scroll_ms Number of milliseconds to wait after scroll
        """
        raise NotImplementedError
    
    def scroll_into_view(self, element, sleep_after_scroll_ms=100):
        """
        Scroll so that given element is entirely visible
        @param element Element to view
        @param sleep_after_scroll_ms Number of milliseconds to wait after scroll
        """
        raise NotImplementedError
    
    def scroll_to(self, x, y, sleep_after_scroll_ms=100):
        """
        Scroll to given coordinates
        @param x X coordinate (pixel number) in top left corner
        @param y y coordinate (pixel number) in top left corner
        @param sleep_after_scroll_ms Number of milliseconds to wait after scroll
        """
        raise NotImplementedError
    
    def wait_until_window_is_loaded(self, timeout_seconds=None):
        """
        Wait until window is loaded with default timeout
        @param timeout_seconds Timeout in seconds
        """
        raise NotImplementedError
    
    
    
    
    


# -*- coding: utf-8 -*-

import logging
from holado_core.common.tools.tools import Tools
from holado_rabbitmq.tools.rabbitmq.rabbitmq_client import RMQBufferConsumer
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_python.standard_library.typing import Typing
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.handlers.wait import WaitEndChange, WaitChange

logger = logging.getLogger(__name__)


class RMQManager:

    def __init__(self):
        self.__protobuf_messages = None
        
    def initialize(self, protobuf_messages):
        self.__protobuf_messages = protobuf_messages
    
    
    def received_messages_as_protobuf_objects(self, consumer, message_type_fullname):
        """Unserialize messages received by a buffer consumer to given Protobuf message type"""
        
        if not isinstance(consumer, RMQBufferConsumer):
            raise TechnicalException(f"Consumer has to be a buffer consumer (obtained type: {Typing.get_object_class_fullname(consumer)})")
        
        res = []
        for ind, m in enumerate(consumer.messages):
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"Received message [{ind}]: {m[3]}")
            msg = self.__protobuf_messages.new_message(message_type_fullname, serialized_string=m[3])
            res.append(msg)
            # logger.debug(f"+++++ New message: type={Typing.get_object_class_fullname(msg)} ; dir={dir(msg)}")
        return res
    
        
    def await_message_reception(self, consumers, timeout_seconds=None, polling_seconds=None, raise_exception=True):
        """Wait until one of the buffer consumers has received a message.
        @param consumers: list of buffer consumers
        @param timeout_seconds: wait timeout
        @param polling_seconds: time to wait between pollings of received messages of all consumers
        """
        if raise_exception is None:
            raise_exception = True
        
        wait = WaitChange(f"Wait message reception by consumers",
                          lambda: sum([c.nb_messages for c in consumers if c.nb_messages > 0]),
                          do_process_in_thread = False,
                          result_reference = 0,
                          timeout_seconds=timeout_seconds, polling_seconds=polling_seconds)
        # On timeout, we will have res==0 and raise is managed before return
        wait.without_raise_on_timeout()
        
        res = wait.execute()
        
        if res == 0 and raise_exception:
            names = ",".join([c.name for c in consumers])
            raise FunctionalException(f"[{names}] No message was received (timeout: {timeout_seconds} seconds)")
        return res
    

    #TODO EKL: add a first_accepted_seconds and a accepted_window_seconds to be able to raise an exception that first or next message arrived but too late
    #TODO EKL: change use of polling_seconds to something like a timer, rather than waiting polling_seconds
    def await_messages_reception(self, consumers, first_timeout_seconds=None, window_seconds=None, polling_seconds=None, raise_exception=True):
        """Wait until buffer consumers stop to receive messages.
        It begins by waiting a first message in any consumer, since their creation or last reset.
        If no consumer receives a first message during first_timeout_seconds period, wait stops.
        Then after each message received by any consumer, a window duration of window_seconds is waited for a new message.
        When no new message is received by any consumer during the window period, wait stops.
        @param consumers: list of buffer consumers
        @param first_timeout_seconds: reception timeout of a first message
        @param window_seconds: time window for wait of a new message since last received message
        @param polling_seconds: time to wait between pollings of received messages of all consumers
        """
        # res = 0
        #
        # if window_seconds is None:
        #     if first_timeout_seconds is not None and first_timeout_seconds > 0:
        #         window_seconds = first_timeout_seconds / 10
        #     else:
        #         window_seconds = 1
        #     if window_seconds < 0.01:
        #         window_seconds = 0.01       # a window period below 10 ms is usually not efficient in testing context
        # if polling_seconds is None:
        #     min_window = min(first_timeout_seconds, window_seconds) if first_timeout_seconds is not None else window_seconds
        #     polling_seconds = min_window / 100       # 1% of error on window detection
        #     if polling_seconds > 0.1:
        #         polling_seconds = 0.1       # a polling period over 100 ms is usually not efficient in testing context
        #     elif polling_seconds < 0.001:
        #         polling_seconds = 0.001       # a polling period below 1 ms is usually not efficient in testing context
        #
        # # Wait first message is needed
        # dt_begin = DateTime.now()
        # dt_last_poll = dt_begin
        # if first_timeout_seconds is not None:
        #     while (dt_last_poll - dt_begin).total_seconds() < first_timeout_seconds:
        #         nb_msg_by_consumer = {c.name: c.nb_messages for c in consumers if c.nb_messages > 0}
        #         dt_last_poll = DateTime.now()
        #         if len(nb_msg_by_consumer) > 0:
        #             res = sum(nb_msg_by_consumer.values())
        #             if Tools.do_log(logger, logging.DEBUG):
        #                 logger.debug(f"Received first messages: total={res} ; " + " ; ".join(f"{k}:{v}" for k,v in nb_msg_by_consumer.items()))
        #             break
        #         time.sleep(polling_seconds)
        #     if res == 0:
        #         if raise_exception:
        #             names = ",".join([c.name for c in consumers])
        #             raise FunctionalException(f"[{names}] No message was received (timeout: {first_timeout_seconds} seconds)")
        #         return res
        #
        # # Wait end of reception
        # dt_last_receive = dt_last_poll
        # while (dt_last_poll - dt_last_receive).total_seconds() < window_seconds:
        #     nb_msg_by_consumer = {c.name: c.nb_messages for c in consumers if c.nb_messages > 0}
        #     nb = sum(nb_msg_by_consumer.values())
        #     dt_last_poll = DateTime.now()
        #     if nb > res:
        #         res = nb
        #         if Tools.do_log(logger, logging.DEBUG):
        #             logger.debug(f"Received messages: total={res} ; " + " ; ".join(f"{k}:{v}" for k,v in nb_msg_by_consumer.items()))
        #         dt_last_receive = dt_last_poll
        #     time.sleep(polling_seconds)
        #
        # return res
        if raise_exception is None:
            raise_exception = True
        
        wait = WaitEndChange(f"Wait end of message reception by consumers",
                             lambda: sum([c.nb_messages for c in consumers if c.nb_messages > 0]),
                             do_process_in_thread = False,
                             result_reference = 0,
                             first_timeout_seconds=first_timeout_seconds, window_seconds=window_seconds, polling_seconds=polling_seconds)
        # On timeout, we will have res==0 and raise is managed before return
        wait.without_raise_on_timeout()
        
        res = wait.execute()
        
        if res == 0 and raise_exception:
            names = ",".join([c.name for c in consumers])
            raise FunctionalException(f"[{names}] No message was received (timeout: {first_timeout_seconds} seconds)")
        return res
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import abc
from holado_core.common.handlers.abstracts.base_redo import BaseRedo

logger = logging.getLogger(__name__)


class Redo(BaseRedo):
    """ Redo a process until timeout.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, name):
        super().__init__(name)
        self.__is_redo_while_null = False
        self.__is_redo_while_not_null = False
        self.__redo_while_values = []
        self.__is_redo_until_none = False
        self.__is_redo_until_not_none = False
        self.__redo_until_values = []
    
    def _is_redo_needed(self, result):
        if self.__is_redo_while_null and result is None:
            return True
        if self.__is_redo_until_none and result is None:
            return False
        
        if self.__is_redo_while_not_null and result is not None:
            return True
        if self.__is_redo_until_not_none and result is not None:
            return False
        
        if len(self.__redo_while_values) > 0 and result is not None:
            for value in self.__redo_while_values:
                if result == value:
                    return True
        
        if len(self.__redo_until_values) > 0:
            if result is None:
                return True
            else:
                for value in self.__redo_until_values:
                    if result == value:
                        return False
        
        # If redo until something, then return True, else return False
        return self.__is_redo_until_none or self.__is_redo_until_not_none or len(self.__redo_until_values) > 0
    
    def redo_while_none(self):
        """
        Redo while result is None
        Note: can be combined with any other redo_XXX method
        @return self
        """
        self.__is_redo_while_null = True
        return self
    
    def redo_while_not_none(self):
        """
        Redo while result is not None
        Note: can be combined with any other redo_XXX method
        @return self
        """
        self.__is_redo_while_not_null = True
        return self
    
    def redo_while(self, value):
        """
        Redo while result is equal to given value
        Note: can be combined with any other redo_XXX method
        @param value Value
        @return self
        """
        self.__redo_while_values.append(value)
        return self
    
    def redo_until_none(self):
        """
        Redo until result is None
        Note: can be combined with any other redo_XXX method
        @return self
        """
        self.__is_redo_until_none = True
        return self
    
    def redo_until_not_none(self):
        """
        Redo until result is not None
        Note: can be combined with any other redo_XXX method
        @return self
        """
        self.__is_redo_until_not_none = True
        return self
    
    def redo_until(self, value):
        """
        Redo until result is equal to given value
        Note: can be combined with any other redo_XXX method
        @param value Value
        @return self
        """
        self.__redo_until_values.append(value)
        return self
        
        
        
